import math 
import numpy as np
from desalsim.density_calc import density_calc 
from desalsim import constants 
from desalsim import scaleup

#%%
#Molecular weight 
MW_Na=constants.MW_Na
MW_Cl=constants.MW_cl
MW_SO4=constants.MW_so4
MW_K=constants.MW_K
MW_Ca=constants.MW_Ca
MW_Mg=constants.MW_Mg
MW_HCO3=constants.MW_HCO3
MW_H=1.00784
MW_OH=17.008
MW_H2O=MW_H+MW_OH

#%%Calculations 
class EDBMCalc:
    """
    A class used to represent Mass Balance for Electrodialysis With Bipolar Membranes 
    ...

    Attributes
    ----------

    CNa_in, CCl_in, CK_in, CMg_in, CCa_in, CSO4_in : float
        Initial concentrations of various ions (g/l).
    CNa_out, CCl_out, CK_out, CMg_out, CCa_out, CSO4_out : float
        Outlet concentrations of various ions (g/l).
    Ci_s_in: float 
        Initial concentrations of various ions in salt channel (mol/l)
    Ci_a_in: float 
        Initial concentrations of various ions in acid channel (mol/l)
    Ci_b_in: float 
        Initial concentrations of various ions in base channel (mol/l)  
    EMF: float
        electromotive force (V) 
    KW_s_in: float
        Inlet ionic water product in salt channel 
    KW_a_in: float
        Inlet ionic water product in acid channel 
    KW_b_in: float
        Inlet ionic water product in base channel         
    M_h2o_a_in: float 
        Initial mass flow rate of water in acid channel (kg/h)
    M_h2o_b_in: float 
        Initial mass flow rate of water in base channel (kg/h)
    M_h2o_s_in: float 
        Initial mass flow rate of water in salt channel (kg/h)        
    N_trip: integer 
        Number of triplets of a channel 
    P: float
        Gross power needed (W)
    PM: float
        Molecular weight
    Q : float
        Flow rate (l/h)
    V_ext: float
        Voltage needed (V) 


    Methods
    -------
    flowrate(): 
        Calculates the flowrate in each channel 
    
    in_mass_flow_rates():
        Calculates the inlet mass flow rates of each ion, kg/h 
    
    acid_channel():
        Calculates flow rates and outlet concentration in Acid channel 
    
    base_channel():
        Calculates flow rates and outlet concentration in Base channel 
        
    Salt_channel():
        Calculates flow rates and outlet concentration in Salt channel 
    """
    def __init__(self, Qin, A, I_d, N, C_s_in, C_b_in, C_a_in, T ):
        #input conditions   
        F=96485.3 #Coulombs/mol
        self.A=A
        self.I_ext=self.A*I_d
        self.JA=3.6*self.I_ext/F
        self.T=T
        self.N_trip=N 
        #Initialization of parameters
        self.Qin=Qin
        
        Cc1, Cc2, Cc3, Cc4, Cc5, Cc6, Cc7, CH, COH= C_s_in
        Cb1, Cb2,Cb3,Cb4,Cb5,Cb6,Cb7,Cb8,Cb9=C_b_in
        Ca1,Ca2,Ca3,Ca4,Ca5,Ca6,Ca7,Ca8,Ca9=C_a_in
        self.CNa_s_in=Cc1/MW_Na #Na concentration (g/l) to (mol/l)
        self.CCl_s_in=Cc2/MW_Cl #Cl concentration (g/l) to (mol/l)
        self.CK_s_in=Cc3/MW_K #K concentration (g/l) to (mol/l)
        self.CMg_s_in=Cc4/MW_Mg #Mg concentration (g/l) to (mol/l)
        self.CCa_s_in=Cc5/MW_Ca #Ca concentration (g/l) to (mol/l)
        self.CSO4_s_in=Cc6/MW_SO4 #SO4 concentration (g/l) to (mol/l)
        self.CHCO3_s_in=Cc6/MW_HCO3 #HCO3 concentration (g/l) to (mol/l)      
        self.COH_s_in=COH/MW_OH
        self.CH_s_in=CH/MW_H
        self.PM_i=[MW_Na, MW_Cl, MW_K, MW_Mg, MW_Ca, MW_SO4, MW_HCO3, MW_H, MW_OH]
        
        self.d_a=density_calc(self.T,sum(C_a_in) )/1000
        self.d_b=density_calc(self.T,sum(C_b_in) )/1000
        self.d_s=density_calc(self.T,sum(C_s_in) )/1000
        
        #Create lists for initial concentration in each channel 
        self.Ci_s_in=[self.CNa_s_in,self.CCl_s_in,self.CK_s_in,self.CMg_s_in,self.CCa_s_in,self.CSO4_s_in,self.CHCO3_s_in, self.CH_s_in,self.COH_s_in]
        self.Ci_b_in=[Cb1/MW_Na, Cb2/ MW_Cl,Cb3/MW_K,Cb4/MW_Mg,Cb5/MW_Ca,Cb6/MW_SO4,Cb7/MW_HCO3,Cb8/MW_H,Cb9/MW_OH]
        self.Ci_a_in=[Ca1/MW_Na,Ca2/MW_Cl,Ca3/MW_K,Ca4/MW_Mg,Ca5/MW_Ca,Ca6/MW_SO4,Ca7/MW_HCO3,Ca8/MW_H,Ca9/MW_OH]
        
        #save initial concentration before recirculation: 
        self.Ci_s_in_0=self.Ci_s_in
        self.Ci_b_in_0=self.Ci_b_in
        self.Ci_a_in_0=self.Ci_a_in

        
    def flowrate(self):
        # Calculates the flowrate in each channel 
        self.Q1_s_in=self.Qin/self.N_trip #units: l/h
        self.Q1_b_in=self.Qin/self.N_trip #units: l/h
        self.Q1_a_in=self.Qin/self.N_trip #units: l/h

    def in_mass_flow_rates(self, ph_s):
        #Calculates the inlet mass flow rates of each ion, kg/h 
        #initialize lists 
        self.M_s_in=[[],[],[],[],[],[],[],[],[]]
        self.M_b_in=[[],[],[],[],[],[],[],[],[]]
        self.M_a_in=[[],[],[],[],[],[],[],[],[]]
                
        for i in range(0,9):
            self.M_s_in[i]=self.Q1_s_in*self.Ci_s_in[i]*self.PM_i[i]/1000 #units: kg/h
            self.M_b_in[i]=self.Q1_b_in*self.Ci_b_in[i]*self.PM_i[i]/1000 #units: kg/h
            self.M_a_in[i]=self.Q1_a_in*self.Ci_a_in[i]*self.PM_i[i]/1000 #units: kg/h
            
        #Calculate mass of water in the initial streams 
        self.M_h2o_a_in=self.Q1_a_in*self.d_a-sum(self.M_a_in)          
        self.M_h2o_b_in=self.Q1_b_in*self.d_b-sum(self.M_b_in) 
        self.M_h2o_s_in=self.Q1_s_in*self.d_s-sum(self.M_s_in) 
        
        #Calculate inlet ionic water product in each channel 
        self.KW_s_in=self.Ci_s_in[7]*self.Ci_s_in[8] 
        self.KW_a_in=self.Ci_a_in[7]*self.Ci_a_in[8]
        self.KW_b_in=self.Ci_b_in[7]*self.Ci_b_in[8]
    
        #Calculation of the inlet concentration of the Cl in salt channel (exception)
        if ph_s<7:
            self.M_s_in[1]=self.Q1_s_in*self.Ci_s_in[1]*self.PM_i[1]/1000+(self.M_s_in[7]/self.PM_i[7]-self.M_s_in[8]/self.PM_i[8])*(self.PM_i[1])
        elif ph_s>7:
            self.M_s_in[1]=self.Q1_s_in*self.Ci_s_in[1]*self.PM_i[1]/1000+(self.M_s_in[8]/self.PM_i[8]-self.M_s_in[7]/self.PM_i[7])*self.PM_i[1]
        

    def acid_channel(self):
        #Performs mass balance calculations for Acid channel 
        #Initialize ionic Mass and concentration in acid channel  
        self.M_a_out=[[],[],[],[],[],[],[],[],[]]
        self.Ci_a_out=[[],[],[],[],[],[],[],[],[]]
        
        #Calculation of outlet mass flow rate for H+
        self.M_a_out[7]=self.M_a_in[7]+self.JA*self.PM_i[7]
        
        #Calculation of outlet mass flow rate for Cl-
        self.M_a_out[1]=self.M_a_in[1]+self.JA*self.PM_i[1]
        #Calculation of outlet mass flow rate for OH-
        self.M_a_out[8]=self.M_a_in[8]
        #Calculation of outlet mass flow rate for Na+
        self.M_a_out[0]=self.M_a_in[0]
        #Calculation of outlet mass flow rate for water 
        self.M_h2o_a_out=self.M_h2o_a_in-0.5*self.JA*MW_H2O
        #Calculation total outlet mass flow rate 
        self.M_a_out_t=self.M_h2o_a_out+self.M_a_out[7]+self.M_a_out[1]+self.M_a_out[8]+self.M_a_out[0]

        #Calculation of outlet mass flow rate for other ionic species in channel 
        for i in range(2,7):
            self.M_a_out[i]=self.M_a_in[i]
            self.M_a_out_t=self.M_a_out_t+self.M_a_out[i]
        
        #Calculate of volumetric outlet flow rate
        self.Q1_a_out=self.M_a_out_t/1
        #Calculation of outlet concentration of single ions in channel 
        for i in range(0,9):
            self.Ci_a_out[i]=self.M_a_out[i]/(self.Q1_a_out*self.PM_i[i]/1000)

        
    def base_channel(self):
        #Performs mass balance calculations for Base channel 
        #Initialize ionic Mass and concentration in base channel  
        self.M_b_out=[[],[],[],[],[],[],[],[],[]]
        self.Ci_b_out=[[],[],[],[],[],[],[],[],[]]
       
        #Calculation of outlet mass flow rate for Na+
        self.M_b_out[0]=self.M_b_in[0]+self.JA*self.PM_i[0]
        #Calculation of outlet mass flow rate for OH-        
        self.M_b_out[8]=self.M_b_in[8]+self.JA*self.PM_i[8]
        #Calculation of outlet mass flow rate for Cl-
        self.M_b_out[1]=self.M_b_in[1]
        #Calculation of outlet mass flow rate for H+
        self.M_b_out[7]=self.M_b_in[7]
        #Calculation of outlet mass flow rate for water 
        self.M_h2o_b_out=self.M_h2o_b_in+0.5*self.JA*MW_H2O
        #Calculation total outlet mass flow rate 
        self.M_b_out_t=self.M_h2o_b_out+self.M_b_out[7]+self.M_b_out[1]+self.M_b_out[8]+self.M_b_out[0]
 
        #Calculation of outlet mass flow rate for other ionic species in channel 
        for i in range(2,7):
            self.M_b_out[i]=self.M_b_in[i]  
            self.M_b_out_t=self.M_b_out_t+self.M_b_out[i]
            
        d_s_new=density_calc(25, self.M_b_out_t)/1000
        #Calculate of volumetric outlet flow rate
        self.Q1_b_out=self.M_b_out_t/d_s_new #assume density is 1kg/l
        
        #Calculation of outlet concentration of single ions in channel 
        for i in range(0,9):
            self.Ci_b_out[i]=self.M_b_out[i]/(self.Q1_b_out*self.PM_i[i]/1000)
    
    def salt_channel(self, Cm_bp_H, Cm_bp_OH):
        #Performs mass balance calculations for Salt channel 
        #Initialize ionic Mass and concentration in salt channel  
        self.M_s_out=[[],[],[],[],[],[],[],[],[]]
        self.Ci_s_out=[[],[],[],[],[],[],[],[],[]]
        
        #Calculation of outlet mass flow rate for Na+
        self.M_s_out[0]=self.M_s_in[0]-(self.M_b_out[0]-self.M_b_in[0])
        #Calculation of outlet mass flow rate for Cl-
        self.M_s_out[1]=self.M_s_in[1]-(self.M_a_out[1]-self.M_a_in[1])
        #Calculation of outlet mass flow rate for OH-   
        self.M_s_out[8]=self.M_s_in[8]
        #Calculation of outlet mass flow rate for H+
        self.M_s_out[7]=self.M_s_in[7]
        #Calculation of outlet mass flow rate for water 
        self.M_h2o_s_out=self.M_h2o_s_in
        #Calculation total outlet mass flow rate 
        self.M_s_out_t=self.M_s_out[0]+self.M_s_out[1]+self.M_h2o_s_out#+self.M_s_out[8]+self.M_s_out[7]
        #Calculation of outlet mass flow rate for other ionic species in channel 
        for i in range(2,7):
            self.M_s_out[i]=self.M_s_in[i]
            self.M_s_out_t=self.M_s_out_t+self.M_s_out[i]

        #Calculate of volumetric outlet flow rate
        self.Q1_s_out=self.M_s_out_t/self.d_s
        
        #Calculation of outlet concentration of single ions in channel 
        for i in range(0,9):
            self.Ci_s_out[i]=self.M_s_out[i]/(self.Q1_s_out*self.PM_i[i]/1000)     

        
        #Calculation of voltage needed and electromotive force 
        self.c1_emf=math.log(((self.Ci_a_in[7]+self.Ci_a_out[7])/2)/Cm_bp_H)        
        self.c2_emf=-math.log(Cm_bp_OH/((self.Ci_b_in[8]+self.Ci_b_out[8])/2))        
        self.c3_emf=-math.log(((self.Ci_s_in[1]+self.Ci_s_out[1])/2)/((self.Ci_a_in[1]+self.Ci_a_out[1])/2))
        self.c4_emf=math.log(((self.Ci_b_in[0]+self.Ci_b_out[0])/2)/((self.Ci_s_in[0]+self.Ci_s_out[0])/2))  
        
            #EMF: electromotive force (V) 
        F=96485.3 #Coulombs/mol
        R_const=8.314462618 #kg⋅m2⋅s−2⋅K−1⋅mol−1
        # R_int=0.28 #ohm cm2
        R_int=45 #ohm cm2
        z=1
        self.EMF=((R_const*self.T/z/F)*(self.c1_emf+self.c2_emf+self.c3_emf+self.c4_emf))*self.N_trip   
    
            #V_ext: Voltage needed (V) 
        self.V_ext=self.EMF+((self.I_ext*R_int)/(self.A*10000))*self.N_trip
        
            #Calculate gross power needed (W)
        self.P=self.V_ext*self.I_ext         

#%% Example usage 
#constants
I_d=400 # The electricl current desnity Am2

F=96485.3 #Coulombs/mol
R_const=8.314462618 #kg⋅m2⋅s−2⋅K−1⋅mol−1
# R_int=0.28 #ohm cm2
R_int=45 #ohm cm2
z=1
npump=0.8 #pump efficiency (units: -)
dp=1 #pressure drop (units: bar)

#Membrane characteristics

Cm_bp_H= 0.0000001 #mol/l 
Cm_bp_OH= 0.0000001 #mol/l 

#Molecular weight 
MW_Na=constants.MW_Na
MW_Cl=constants.MW_cl
MW_SO4=constants.MW_so4
MW_K=constants.MW_K
MW_Ca=constants.MW_Ca
MW_Mg=constants.MW_Mg
MW_HCO3=constants.MW_HCO3
MW_H=1.00784
MW_OH=17.008
MW_H2O=MW_H+MW_OH

T=20+273.15

#input data
r=0.91 #recycling rate 
rs=0.75 #recycling rate 
ph_s=4.71 #pH salt channel (units: -)
ph_b=7#pH base channel (units: -)
ph_a=7#pH acid channel (units: -)
d_a=0.997 #density acid channel (units: kg/l)
#d_s=1 #density salt channel (units: kg/l)
d_b=0.997 #density base channel (units: kg/l)


#Feed concentration g/L
Cin_edbm=[13.44, 20.725, 1.146, 0, 0, 0.18, 0, 10**(-ph_s), 3.01551E-11] 
d_in=density_calc(25,sum(Cin_edbm))/1000
d_s=d_in

C_b_in=[0,0,0,0,0,0,0,10**(-ph_b), 10**(-(14-ph_b))]
C_a_in=[0,0,0,0,0,0,0,10**(-ph_a), 10**(-(14-ph_a))]

#Feed flow rate L/h
Q_in_edbm=47000

#Calculate water quantity in inflow 
Mw_in=Q_in_edbm/d_in 

#Set number of triplets 
N_trip=50*47 # Number of triplets based on the inlet flow rate

#Set membrane area based on the feed flow rate, m2 
A=0.4 #range: 0.1-1

#Initialize concentration of Na in salt channel 
Cna_s=[]

#Create an instance of the EDBMCalc class with the defined parameters
edbm_dat=EDBMCalc(Q_in_edbm, A, I_d, N_trip, Cin_edbm, C_b_in, C_a_in, T )

# Call the necessary methods to calculate values
edbm_dat.flowrate()
edbm_dat.in_mass_flow_rates(ph_s)
edbm_dat.acid_channel()
edbm_dat.base_channel()
edbm_dat.salt_channel(Cm_bp_H, Cm_bp_OH)
Cna_s.append(edbm_dat.Ci_s_out[0])

# # Sum results

"Salt channel "
    #Concentration in salt channel 
Cbrine_out_t=sum(edbm_dat.Ci_s_out)
Cbrine_out=edbm_dat.Ci_s_out #mol/l
Cbrine_out_g=[Cbrine_out[0]*MW_Na, Cbrine_out[1]*MW_Cl, Cbrine_out[2]*MW_K, Cbrine_out[3]*MW_Mg, Cbrine_out[4]*MW_Ca, Cbrine_out[5]*MW_SO4] #g/l

     #Mass flow rate
M_s_out=edbm_dat.M_s_out_t*N_trip
print("Salt channel: Mass flow rate out is "+str(round(M_s_out,2))+"kg/hr")
    #Volumetric flow rate 
Q_s_out=edbm_dat.Q1_s_out*N_trip
print("Salt channel: Volumetric flow rate out is "+str(round(Q_s_out,2))+"l/hr")
print("Na concentration:"+str(round(Cbrine_out[0],2))+"M and "+str(round(Cbrine_out_g[0],2))+"g/l")
print("Cl concentration:"+str(round(Cbrine_out[1],2))+"M and "+str(round(Cbrine_out_g[1],2))+"g/l")
print("-----------------------------------------")

"Base channel "
    #Concentration in base channel 
Cb_out=edbm_dat.Ci_b_out[0:6]
Cb_out_g=[Cb_out[0]*MW_Na, Cb_out[1]*MW_Cl, Cb_out[2]*MW_K, Cb_out[3]*MW_Mg, Cb_out[4]*MW_Ca, Cb_out[5]*MW_SO4]

     #Mass flow rate
M_b_out=edbm_dat.M_b_out_t*N_trip
print("Base channel: Mass flow rate out is "+str(round(M_b_out,2))+"kg/hr")

     #Volumetric flow rate 
Q_b_out=edbm_dat.Q1_b_out*N_trip
print("Base channel: Volumetric flow rate out is "+str(round(Q_b_out,2))+"l/hr")

print("Na concentration "+str(round(Cb_out[0],2))+"M and "+str(round(Cb_out_g[0],2))+"g/l")
    #Conversion to solid 
M_NaOH_out=Q_b_out*edbm_dat.Ci_b_out[0]*constants.MW_NaOH/1000 #kg/hr 
print("-----------------------------------------")

"Acid channel" 
    #Concentration in acid channel 
Ca_out=edbm_dat.Ci_a_out
Ca_out=edbm_dat.Ci_a_out[0:6]
Ca_out_g=[Ca_out[0]*MW_Na, Ca_out[1]*MW_Cl, Ca_out[2]*MW_K, Ca_out[3]*MW_Mg, Ca_out[4]*MW_Ca, Ca_out[5]*MW_SO4]

    #Mass flow rate 
M_a_out=edbm_dat.M_a_out_t*N_trip
print("Acid channel: Mass flow rate out is "+str(round(M_a_out,2))+"kg/hr")

    #Volumetric flow rate 
Q_a_out=edbm_dat.Q1_a_out*N_trip
print("Acid channel: Volumetric flow rate out is "+str(round(Q_a_out,2))+"l/hr")
print("Cl concentration "+str(round(Ca_out[1],2))+"M and "+str(round(Ca_out_g[1],2))+"g/l")
print("-----------------------------------------")

    #Conversion to solid 
M_HCl_out=Q_a_out*constants.MW_HCl/1000 #kg/hr

#Calculate required amount of water for the operation mode 
Q_w_in=2*Q_in_edbm

#Calculate mass balance 
bal=(Q_in_edbm*d_in+2*Q_in_edbm)-M_a_out-M_s_out-M_b_out
print("Mass balance difference is "+str(round(bal,2)))
error_perc=abs(bal)/(Q_in_edbm*d_in+2*Q_in_edbm)*100
print("Balance error percentage is "+str(round(error_perc,2))+"%")

#Energy consumption 
V_ext=edbm_dat.V_ext #xternal 
#Calculate energy consumption for pumping 
Ppump=(edbm_dat.Q1_s_in*N_trip*dp+edbm_dat.Q1_a_in*N_trip*dp+edbm_dat.Q1_b_in*N_trip*dp)/1000/3600*1e5/npump #units: W -> l to m3 so /1000; bar to J 1e5N/m2*1J/m ; hr to 3660s  

#Calculate current efficiency 
Cb_in=[0]
CE=(Q_b_out)*(Cb_out[0]-Cb_in[0])*F/(3600*N_trip*I_d*A)*100 #%
print("Current efficiency is "+str(round(CE,2))+"%")
print("-----------------------------------------")
#Total energy consumption 
E_el_Edbm=V_ext*I_d*A/1000+Ppump/1000
SEC=(V_ext*I_d*A)/(Q_b_out*(edbm_dat.Ci_b_out[0]-edbm_dat.Ci_b_in[0])*constants.MW_NaOH)

print("Total electrical consumption for EDBM is " + str(round(E_el_Edbm,2))+ " KW")
print("Specific energy consumption is "+str(round(SEC,2))+"kwh/kg NaOH")


import math
from desalsim import constants 
from desalsim.density_calc import density_calc 
#%%
#%%constants
R=8.31446261815324 #gas constant
T=20+273.15 #K
MW_Na=22.99
MW_cl=35.453
MW_so4=32.064+4*15.999
MW_K=39.102
MW_Ca=40.08
MW_Mg=24.31
MW_HCO3=1.008+12.011+3*15.999
#assumptions 
dp=2

#molarity 
class molarity:
    def __init__(self, MW, zi, Ci):
        self.MW=MW
        self.zi=zi
        self.Ci=Ci
        self.meq=self.Ci*1000*self.zi/self.MW
        #self.mi=self.Ci/(self.MW*(1000-self.sum_conc)/1000                    #CHECK AGAIN THE UNITS 
        
#%%
class nfmass:
    def __init__(self, comp, Cfeedi, rjr, Wrec, Qf):
        self.comp=comp
        self.Cfeedi=Cfeedi #ion concentration g/l
        self.rjr=rjr #rejection of each ion of the membrane 
        self.Wrec = Wrec # water recovery first pass 
        self.Qf=Qf
        
                       
    def permcal(self):
        self.Qperm= self.Wrec*self.Qf
        self.Qconc=self.Qf-self.Qperm
        self.Cpermi = (1-self.rjr)*self.Cfeedi
        self.Cconci = (self.Qf*self.Cfeedi-self.Qperm*self.Cpermi)/self.Qconc
#        print(" ion concenctration of " + self.comp+ " in permeate stream is"+" "+ str(self.Cpermi))
#        print(" ion concenctration of " + self.comp+ " in concentrate stream is"+" "+str(self.Cconci))

#%% osmotic pressure 
class Osmotic_pressure:
    def __init__(self, C1, z1,  C2, z2,  C3, z3,  C4, z4, C5, z5, C6, z6):
        self.C1=C1
        self.z1=z1
        self.C2=C2
        self.z2=z2
        self.C3=C3
        self.z3=z3
        self.C4=C4
        self.z4=z4
        self.C5=C5
        self.z5=z5
        self.C6=C6
        self.z6=z6
        self.sum_Ci=sum([C1, C2, C3, C4, C5, C6])
        self.zi_2=[z1**2, z2**2, z3**2, z4**2, z5**2, z6**2]
        self.conc=[C1/MW_Na, C2/MW_cl, C3/MW_K, C4/MW_Mg, C5/MW_Ca, C6/MW_so4] #mol/l
        self.mi=[C1*1000/(MW_Na*1000*((1e+6-self.sum_Ci*1000)/1e+6)), C2*1000/(MW_cl*1000*((1e+6-self.sum_Ci*1000)/1e+6)), C3*1000/(MW_K*1000*((1e+6-self.sum_Ci*1000)/1e+6)), C4*1000/(MW_Mg*1000*((1e+6-self.sum_Ci*1000)/1e+6)), C5*1000/(MW_Ca*1000*((1e+6-self.sum_Ci*1000)/1e+6)), C6*1000/(MW_so4*1000*((1e+6-self.sum_Ci*1000)/1e+6))]
        self.mizi_2=[]
        
    # def osmotic_pressure_calculation(self):
        for i in range(6) :
            self.mizi_2.append(self.mi[i]*self.zi_2[i])
        self.SI=sum(self.mizi_2)/2
        self.B=-348.662/(T)+6.72817-0.971307*math.log(T)
        self.C=40.5016/(T)-0.721404+0.103915*math.log(T)
        self.D=5321/(T)+233.76-0.9297*(T)+0.001417*(T)**2-0.0000008292*(T)**3
        self.S=1.17202*(sum(self.mizi_2)/sum(self.mi))*0.9982**0.5*(23375.556/(self.D*(T)))**1.5
        self.fi=1-self.S/(3.375*self.SI)*((1+1.5*self.SI**0.5)-2*math.log(1+1.5*self.SI**0.5)-1/(1+1.5*self.SI**0.5))+self.B*sum(self.mi)/2+self.C*(sum(self.mi)/2)**2
        self.sum_conc=sum(self.conc)
        self.p_osmo_psig=1.205*self.fi*(T)*sum(self.mi)
        self.p_osmo=self.p_osmo_psig/14.3 #bar
            
#%%Energy consumption 
class NfEnergy:   
    def __init__(self, P_osmo_c, P_osmo_f, P_osmo_p, dp, d_p, Qperm, Qf, d_in):
        self.P_osmo_c = P_osmo_c # Osmotic pressure of concentrate stream (units:bar)        
        self.P_osmo_f = P_osmo_f # Osmotic pressure of feed stream (units:bar)
        self.P_osmo_p = P_osmo_p # Osmotic pressure of permeate stream (units:bar)
        self.dp = dp #pressure drop (units:bar)
        self.d_p = d_p #Permeate stream density 
        self.Qperm = Qperm # Permeate flow rate
        self.Qf = Qf # Concentrate flow rate
        self.d_in = d_in #Feed stream density 

    def calculate_energy_consumption(self):
        Papplied = (self.P_osmo_c + self.P_osmo_f) / 2 - self.P_osmo_p + self.dp #Applied pressure (units: bar)
        Ppump = Papplied * self.Qperm / self.d_p * 1e5 / 3600 #Power for applied pressure (units:W)
        E_el_nf = (Ppump / 1000 / 0.8 ) #Electricity consumption (units:KW)
        spec = E_el_nf / (self.Qperm / self.d_p) #Specific energy consumption Kwh/m3 of permeate
        SEC_el_feed = E_el_nf / (self.Qf / self.d_in) #Specific energy consumption Kwh/m3 of feed
        
        return {
    "Applied pressure": Papplied,
    "Power for pump": Ppump,
    "E_el_nf": E_el_nf,
    "Spec": spec,
    "SEC_el_feed": SEC_el_feed
}



import unittest
from desalsim.thermal_cryst_f import thermal_calc, conc_cal, calculate_energy
from desalsim.density_calc import density_calc
from desalsim import constants
from desalsim import scaleup

class TestThermalCrystalUnit(unittest.TestCase):
    def setUp(self):
        # Define input parameters for testing
        self.T_op = 60  # Operating temperature (°C)
        self.Qf = 1000  # Inlet flow rate (kg/hr)
        self.Cf_s = 200  # Total ion concentration of solution (g/L)
        Cf_caso4 = 10  # Concentration of CaSO4 in the solution (g/L)
        self.T_in = 40  # Inlet feed temperature (°C)
        Cf_in = [80.42, 116.69, 2.29, 0.01, 0.04, 0.54]  # List of concentration of ions in the solution (g/L)
        self.salt_mois = 20  # Weight percent of NaCl at saturation
        self.LHV_v = 2199.7  # Boiling point elevation (°C)
        self.LHV_s = 2357.69  # Steam temperature (°C)
        self.T_cw_o = 40  # Cooling water inlet temperature (°C)
        self.T_cw_f = 25  # Cooling water outlet temperature (°C)
        self.d_sol=density_calc(self.T_in, sum(Cf_in))/1000 
        
        # Create an instance of the thermal_calc class with the defined parameters
        self.thermal_calc_instance = thermal_calc(self.T_op, self.Qf, self.Cf_s, Cf_caso4,self.T_in, Cf_in, self.salt_mois, self.LHV_v, self.LHV_s,self.T_cw_o, self.T_cw_f)
        self.result_mass=self.thermal_calc_instance.mass_bal_cryst()
        self.result_heat=self.thermal_calc_instance.heat_bal_cryst()
        
        
    def test_mass_bal_cryst(self):
        # Call the mass_bal_cryst method
        self.thermal_calc_instance.mass_bal_cryst()
        self.result_mass=self.thermal_calc_instance.mass_bal_cryst()
        
        self.ev_mass= self.thermal_calc_instance.ev_mass
        self.solid_mass=self.thermal_calc_instance.solid_mass
        
        expected_ev_mass = self.Qf-self.solid_mass
        expected_solid_mass=(self.Qf*self.Cf_s/self.d_sol/1000)/(1-self.salt_mois/100)
        
        # Check if the calculated values are reasonable
        self.assertAlmostEqual(self.solid_mass, expected_solid_mass, delta=expected_solid_mass*0.01)
        self.assertAlmostEqual(self.ev_mass, expected_ev_mass, delta=expected_ev_mass*0.01)

    def test_heat_bal_cryst(self):
        result_mass=self.thermal_calc_instance.mass_bal_cryst()
        
        # Call the heat_bal_cryst method
        Cp_f=3.14 # Feed specific heat capacity (units: KJ* Kg*oC)
        CP_cw=4.187 # Water specific heat capacity (units: KJ* Kg*oC)
        UA=45990
        self.thermal_calc_instance.heat_bal_cryst()
        self.ev_mass= self.thermal_calc_instance.ev_mass
        self.result_heat=self.thermal_calc_instance.heat_bal_cryst()
        
        heat_req=self.thermal_calc_instance.heat_req
        cw_mass=self.thermal_calc_instance.cw_mass
        expected_heat_req=self.LHV_v*self.ev_mass+self.Qf*Cp_f*(self.T_op-self.T_in)
        expected_cw_mass=(self.ev_mass*self.LHV_v)/(CP_cw*(self.T_cw_o-self.T_cw_f)) 
        self.assertAlmostEqual(heat_req, expected_heat_req, delta=expected_heat_req*0.01)
        self.assertAlmostEqual(cw_mass, expected_cw_mass, delta=expected_cw_mass*0.01)

    def test_calculate_energy(self):
        # Define input parameters for calculate_energy function
        Qf = 1000  # Flow rate of the feed solution (m3/h)
        Q_evap_mass = 500  # Distillate flow rate (m3/h)
        Qcw = 200  # Flow rate of cooling water (m3/h)
        M_Nacl = 100  # Mass of NaCl (kg)
        heat_req = 5000  # Heat required (kJ/h)
        d_sol = 1.2  # Density of the feed solution (kg/m3)
        dp_f = 0.1  # Pressure drop for feed (bar)
        dp_w = 1  # Pressure drop for water streams (bar)
        dp_slurry = 3.5  # Pressure drop for slurry streams (bar)
        dp_cw = 2  # Pressure drop for cooling water (bar)
        npump = 0.8  # Pump efficiency

        # Call the calculate_energy function
        E_el_th_Cr, E_th_th_Cr, SEC_el_f, SEC_el_NaCl, SEC_el_w = calculate_energy(Qf, Q_evap_mass, Qcw, M_Nacl,
                                                                                          heat_req, d_sol, dp_f, dp_w,
                                                                                          dp_slurry, dp_cw, npump)
        expected_E_el_th_Cr=0.061
        # Check if the calculated energy values are reasonable
        self.assertAlmostEqual(E_el_th_Cr, expected_E_el_th_Cr, delta=expected_E_el_th_Cr*0.1)


if __name__ == '__main__':
    unittest.main()



import math
import numpy as np
from desalsim.density_calc import density_calc 
from desalsim import constants 
import math
#
    #molecular weight
MW_Na=constants.MW_Na
MW_Cl=constants.MW_cl
MW_SO4=constants.MW_so4
MW_K=constants.MW_K
MW_Ca=constants.MW_Ca
MW_Mg=constants.MW_Mg
MW_HCO3=constants.MW_HCO3
MW_values = [MW_Na, MW_Cl, MW_K, MW_Mg, MW_Ca, MW_SO4]

#%%
class NFMass:
    """
    A class used to represent Mass Balance for Nanofiltration Unit

    ...

    Attributes
    ----------
    comp : str
        Component name
    Cfeedi : float
        Ion concentration in the feed (g/L)
    rjr : float
        Rejection of the ion by the membrane
    Wrec : float
        Water recovery in the first pass
    Qf : float
        Feed flow rate (kg/h)
    Qperm : float
        Permeate flow rate (kg/h)
    Qconc : float
        Concentrate flow rate (kg/h)
    Cpermi : float
        Ion concentration in the permeate (g/L)
    Cconci : float
        Ion concentration in the concentrate (g/L)

    Methods
    -------
    calculate_perm():
        Calculates the permeate and concentrate flow rates, and their ion concentrations
    """

    def __init__(self, comp, Cfeedi, rjr, Wrec, Qf):
        self.comp = comp
        self.Cfeedi = Cfeedi
        self.rjr = rjr
        self.Wrec = Wrec
        self.Qf = Qf
        self.calculate_perm()

    def calculate_perm(self):
        self.Qperm = self.Wrec * self.Qf
        self.Qconc = self.Qf - self.Qperm
        self.Cpermi = (1 - self.rjr) * self.Cfeedi
        self.Cconci = (self.Qf * self.Cfeedi - self.Qperm * self.Cpermi) / self.Qconc

#%% osmotic pressure 
class OsmoticPressure:
    """
        Calculate osmotic pressure for a solution.

         Attributes
         ----------
            C1, C2 C3, C4, C5, C6 : float
                Concentration of ions in the solution (mol/L).
            z1, z2,z3, z4, z5, z6 : int
                Charge of ions in the solution.
        Methods
        -------    
        osmotic_pressure_calculation():
            Calculates the osmotic pressure of a solution.
        Returns
        ------- 
            p_osmo : float
                Osmotic pressure of a solution (bar). 
    """
    def __init__(self, C_values, z_values, T):
        self.Ci = C_values
        self.z = z_values
        self.T = T
        self.sum_Ci=sum(self.Ci)

    def osmotic_pressure_calculation(self):
            """
            Calculates the osmotic pressure of a solution.
            """
            mi=[self.Ci[0]*1000/(MW_Na*1000*((1e+6-self.sum_Ci*1000)/1e+6)), self.Ci[1]*1000/(MW_Cl*1000*((1e+6-self.sum_Ci*1000)/1e+6)), self.Ci[2]*1000/(MW_K*1000*((1e+6-self.sum_Ci*1000)/1e+6)), self.Ci[3]*1000/(MW_Mg*1000*((1e+6-self.sum_Ci*1000)/1e+6)), self.Ci[4]*1000/(MW_Ca*1000*((1e+6-self.sum_Ci*1000)/1e+6)), self.Ci[5]*1000/(MW_SO4*1000*((1e+6-self.sum_Ci*1000)/1e+6))]
            self.mizi_2=[]
            for i in range(6) :
                self.mizi_2.append(mi[i]*self.z[i]**2)
            SI = sum(self.mizi_2) / 2
            B = -348.662 / self.T + 6.72817 - 0.971307 * math.log(self.T)
            C = 40.5016 / self.T - 0.721404 + 0.103915 * math.log(self.T)
            D = 5321 / self.T + 233.76 - 0.9297 * self.T + 0.001417 * self.T ** 2 - 0.0000008292 * self.T ** 3
            S = 1.17202 * (sum(self.mizi_2) / sum(self.Ci)) * 0.9982 ** 0.5 * (23375.556 / (D * self.T)) ** 1.5
            fi=1-S/(3.375*SI)*((1+1.5*SI**0.5)-2*math.log(1+1.5*SI**0.5)-1/(1+1.5*SI**0.5))+B*sum(mi)/2+C*(sum(mi)/2)**2

            #Calculate Osmotic pressure and convert units from psi to bar            
            p_osmo= 1.205 * fi * self.T * sum(mi)/ 14.3 
            return p_osmo
            
#%%Energy consumption 
class NfEnergy:   
    """
        Calculate energy consumption for nanofiltration.
        
            Attributes
            ----------
            P_osmo_c : float
                Osmotic pressure of concentrate stream (bar).
            P_osmo_f : float
                Osmotic pressure of feed stream (bar).
            P_osmo_p : float
                Osmotic pressure of permeate stream (bar).
            dp : float
                Pressure drop (bar).
            d_p:  float
                Permeate stream density (kg/m3).
            Qperm : float
                Permeate flow rate(kg/h).
            Qf : float
                Concentrate flow rate (kg/h).
            d_in : float
                Feed stream density(kg/m3).
            n : float
                Pump efficiency(-).
            Methods
            -------
             calculate_energy_consumption:  
                Calculates the Energy consumption of the nanofiltration unit and the specific energy consumption  
            Returns
            -------
                Papplied : float
                    Applied pressure (bar) 
                Ppump : float
                    Power for applied pressure (W)
                E_el_nf : float
                    Electricity consumption KW
                Spec : float
                    Specific Energy Consumption (Kwh/m3 of permeate)
                SEC_el_feed : float
                    Specific Energy Consumption (Kwh/m3 of feed)
        """
    def __init__(self, P_osmo_c, P_osmo_f, P_osmo_p, dp, d_p, Qperm, Qf, d_in, n):
        self.P_osmo_c = P_osmo_c         
        self.P_osmo_f = P_osmo_f 
        self.P_osmo_p = P_osmo_p 
        self.dp = dp 
        self.d_p = d_p
        self.Qperm = Qperm 
        self.Qf = Qf 
        self.d_in = d_in
        self.n =n
        self.calculate_energy_consumption()

    def calculate_energy_consumption(self):
        """
        Calculate the energy consumption of the nanofiltration unit.

        Returns:
            dict: A dictionary containing the calculated values.
        """
        # Calculate the applied pressure (bar)
        Papplied = (self.P_osmo_c + self.P_osmo_f) / 2 - self.P_osmo_p + self.dp 

        # Calculate the power for the pump
        Ppump = Papplied * self.Qperm / self.d_p * 1e5 / 3600 

        # Calculate the electrical energy consumption (KWh)
        self.E_el_nf = (Ppump / 1000 / self.n ) 

        # Calculate the specific energy consumption for the permeate (KWh/m3 of permeate)
        Spec = self.E_el_nf / (self.Qperm / self.d_p) 

        # Calculate the specific energy consumption for the feed (KWh/m3 of feed)
        SEC_el_feed = self.E_el_nf / (self.Qf / self.d_in) 

        return {
            "Applied pressure (Bar)": round(Papplied,2),
            "Power for pump (KW)": round(Ppump/1000,2) ,
            "E_el_nf (KW)": round(self.E_el_nf,2),
            "Specific Energy Consumption (KWh/m3 of permeate)": round(Spec,2),
            "Specific Energy Consumption (KWh/m3 of feed)": round(SEC_el_feed,2)
        }
#%%
#Example usage
#Feed concentration
components = ['Na', 'Cl', 'K', 'Mg', 'Ca', 'SO4']
Ci_in = [12.33, 21.67, 0.45, 1.39, 0.45, 3.28]
z_values = [1, -1, 1, 2, 2, -2]


#Constants
R=8.314 #gas constant (units: J / mol·K)
T=20+273 #Operating temperature (units: K)

    #molecular weight
MW_Na=constants.MW_Na
MW_Cl=constants.MW_cl
MW_SO4=constants.MW_so4
MW_K=constants.MW_K
MW_Ca=constants.MW_Ca
MW_Mg=constants.MW_Mg
MW_HCO3=constants.MW_HCO3
MW_values = [MW_Na, MW_Cl, MW_K, MW_Mg, MW_Ca, MW_SO4]
mg_in = sum(Ci_in)
#Feed flow density 
d_in = density_calc(T-273, mg_in)  # kg/m3

#Feed flowrate
Qsw = 1000#3000 / 24 * d_in #m3/d
Qf = Qsw  # kg/hr

#Asuumptions  
rjr_values = [0.16, 0.29, 0.21, 0.98, 0.95, 0.98] #Ions rejection rates based on membrane characteristics (units: -)
Wrec = 0.7 # Water recovery based on membrane characteristics (units: -)
n=0.8 #pump efficiency (units: -)
dp=2 # pressure drop (units: bar)


# Function to create NFMass objects for different components
def create_nfmass_objects(components, C_in, rjr_values, Wrec, Qf):
    return [NFMass(comp, Ci, rjr, Wrec, Qf) for comp, Ci, rjr in zip(components, C_in, rjr_values)]

# Create NFMass objects for different components
nfmass_objects = create_nfmass_objects(components, Ci_in, rjr_values, Wrec, Qf)

Cconc = [nf_mass.Cconci for nf_mass in nfmass_objects]
Cperm = [nf_mass.Cpermi for nf_mass in nfmass_objects]
Qperm = nfmass_objects[0].Qperm  # kg/hr
Qconc = nfmass_objects[0].Qconc  # kg/hr
print("Permeate stream flow rate is "+str(round(Qperm,2))+"kg/hr")
print("Permeate stream total concentration is "+str(round(sum(Cperm),2))+"g/l")
print("-----------------------------------------")
print("Concentrate stream flow rate is "+str(round(Qconc,2))+"kg/hr")
print("Concentrate stream total concentration is "+str(round(sum(Cconc),2))+"g/l")
print("-----------------------------------------")

# Calculate Osmotic Pressure
P_osmo_f = OsmoticPressure(Ci_in, z_values, T).osmotic_pressure_calculation()
P_osmo_p = OsmoticPressure(Cperm, z_values, T).osmotic_pressure_calculation()
P_osmo_c = OsmoticPressure(Cconc, z_values, T).osmotic_pressure_calculation()

d_p=density_calc(T-273, sum(Cperm))

#Calculate Energy consumption 
elec=NfEnergy(P_osmo_c, P_osmo_f, P_osmo_p, dp, d_p, Qperm, Qf, d_in,n)
result=elec.calculate_energy_consumption()
for key, value in result.items():
        print(f"{key}: {value}")


import math
from desalsim import scaleup
from desalsim.density_calc import density_calc 
import math
from desalsim import constants 
#%%
    #molecular weight
MW_Na=constants.MW_Na
MW_Cl=constants.MW_cl
MW_SO4=constants.MW_so4
MW_K=constants.MW_K
MW_Ca=constants.MW_Ca
MW_Mg=constants.MW_Mg
MW_HCO3=constants.MW_HCO3
MW_values = [MW_Na, MW_Cl, MW_K, MW_Mg, MW_Ca, MW_SO4]
MW_MgOH=constants.MW_MgOH
MW_CaOH=constants.MW_CaOH

#%%
class MFPFRCALC:
    """
    A class used to represent Mass Balance for percipitation step with the addition of alkaline solution 

    Attributes
    ----------
    MW : float
        Molecular weight of the solute (g/mol)
    Ci_in : float
        Initial concentration of the solute (g/L)
    conv_1: float
        Conversion rate for Magnisium percipitation in the first step 
    conv_2: float
        Conversion rate for Calcium percipitation in the second step    
    C_NaOH_1: float
        Concentration of NaOH solution for first step (mol/L)
    Qin
    QMg_in 
    QNaOH_1 : float 
        Volumetric flow rate of sodium hydroxide in step 1 in L/h
    M_MgOH2_1: float 
        Outlet mass flow rate of magnesium hydroxide produced in kg/h
    Qtot_out_1 : float 
        Outlet volumetric flow rate in L/h
    Mtot_out_1 : float 
        Step 1 outlet mass flow rate in kg/h
    magma_d_1 : float 
        magma density: the quantity of solids produced per volume of slurry kg/l
    ph_1 : float 
        ph of solution during first step 
    kps_MgOH : float 
        Product solubility of Mg(OH)2 
    Ci_out_1 : float 
        The outlet ion concentration from step 1 in mol/L
    QNaOH_2_st : float 
        The stoichiometric volumetric flow rate of sodium hydroxide for the second step L/hr
    QNaOH_2_add : float
        The added volumetric flow rate of sodium hydroxide needed to reach a pH = 13  L/h
    M_CaOH2_2: float 
        The outlet mass flow rate of calcium hydroxide produced during the 2nd step kg/hr 
    M_MgOH2_2 : float 
        The outlet mass flow rate of magnesium hydroxide produced during the 2nd step kg/hr 
    magma_d_2 : float    
        Magma density: the quantity of solids produced per volume of slurry kg/l
    Qtot_out_2 : float 
        Total outlet volumetric flow rate for 2nd step in L/h
    Ci_out_2 : float 
        The outlet ion concentration from step 2 in mol/L
      
    Methods
    -------
    calc_step1():
        Calculates the flowrates, the concentration of the streams and the requirements of alkaline solution in the 1st step 
    calc_step2():
        Calculates the flowrates, the concentration of the streams and the requirements of alkaline solution in the 2nd step 
        """
    def __init__(self, Qin, Cin_mfpfr, C_NaOH_1, C_NaOH_2, conv_1, conv_2):
        self.Qin=Qin
        Cc1, Cc2, Cc3, Cc4, Cc5, Cc6=Cin_mfpfr
        self.CNa_in=Cc1/MW_Na #Na concentration (g/l) to (mol/l)
        self.CCl_in=Cc2/MW_Cl #Cl concentration (g/l) to (mol/l)
        self.CK_in=Cc3/MW_K #K concentration (g/l) to (mol/l)
        self.CMg_in=Cc4/MW_Mg #Mg concentration (g/l) to (mol/l)
        self.CCa_in=Cc5/MW_Ca #Ca concentration (g/l) to (mol/l)
        self.CSO4_in=Cc6/MW_SO4 #SO4 concentration (g/l) to (mol/l)
        self.C_NaOH_1=C_NaOH_1 #Molar concentration of NaOH (mol/l) used in the first precipitation step 
        self.C_NaOH_2=C_NaOH_2 #Molar concentration of NaOH (mol/l) used in the first precipitation step 
        self.conv_1=conv_1 #Conversion rate of Mg the first precipitation step 
        self.conv_2=conv_2 #Conversion rate of Mg the first precipitation step 
        self.d_in=density_calc(25, sum(Cin_mfpfr))
    
    def calc_step1(self, kps_MgOH, d_mgoh_2):
        #Calculate the molar flow rate of magnesium in the reactor during the 1° stepin mol/h
        self.QMg_in=self.Qin*self.CMg_in 
        
        #Calculate the volumetric flow rate of sodium hydroxide in L/h
        self.QNaOH_1=(self.Qin*self.CMg_in*(self.conv_1/100)*2)/self.C_NaOH_1 
        
        #Calculate outlet mass flow rate of magnesium hydroxide produced in kg/h
        self.M_MgOH2_1=(self.Qin*self.CMg_in*(self.conv_1/100)*MW_MgOH)/1000 
        
        #Calculate outlet volumetric flow rate in L/h
        self.Qtot_out_1=self.Qin+self.QNaOH_1
        
        #Calculate outlet mass flow rate in kg/h
        self.Mtot_out_1=self.Qin*self.d_in+self.QNaOH_1*1.04-self.M_MgOH2_1  
        
        #Calculate magma density: the quantity of solids produced per volume of slurry kg/l
        self.magma_d_1=self.M_MgOH2_1/self.Qtot_out_1 
        
        #Calculate ph of solution during the first step 
        self.ph_1=14+math.log10(2*(kps_MgOH/4)**(1/3))
        
        #Calculate total outlet volumetric flow rate in L/h
        self.Qtot_out_1=self.Qin+self.QNaOH_1-self.M_MgOH2_1/d_mgoh_2 
        
        #Calculates the outlet ion concentration in mol/L
        self.CMg_out_1=(self.QMg_in*(1-self.conv_1/100))/self.Qtot_out_1 
        self.CNa_out_1=(self.Qin*self.CNa_in+self.QNaOH_1*self.C_NaOH_1)/self.Qtot_out_1
        self.CCl_out_1=(self.Qin*self.CCl_in)/self.Qtot_out_1
        self.CK_out_1=(self.Qin*self.CK_in)/self.Qtot_out_1
        self.CCa_out_1=(self.Qin*self.CCa_in)/self.Qtot_out_1
        self.CSO4_out_1=(self.Qin*self.CSO4_in)/self.Qtot_out_1
        
        return self.QMg_in, self.QNaOH_1, self.M_MgOH2_1, self.Qtot_out_1, self.Mtot_out_1, self.magma_d_1, self.ph_1, self.CMg_out_1, self.CNa_out_1, self.CCl_out_1, self.CK_out_1, self.CCa_out_1, self.CSO4_out_1
    
    def calc_step2(self, d_mgoh_2, d_caoh_2 ):
       #Calculate the molar flow rate of calcium in the reactor during the 2nd stepin mol/h
       self.QCa_in_2=self.Qtot_out_1*self.CCa_out_1 
       
       #Calculate the stoichiometric volumetric flow rate of sodium hydroxide for the second step L/hr
       self.QNaOH_2_st=(self.Qtot_out_1*(self.CCa_out_1*(self.conv_2/100)+self.CMg_out_1*1)*2)/self.C_NaOH_2 
       
       #Calculate concentration of the hydroxide ion in mol/L for a ph=13 solution  
       self.COH_ph13=10**(-(14-13)) 
       self.COH_st=10**(-(14-self.ph_1))

       #Calculate the added volumetric flow rate of sodium hydroxide needed to reach a pH = 13  L/h
       self.QNaOH_2_add=((0.0216-self.COH_ph13)*(self.QNaOH_2_st+self.Qtot_out_1))/(self.COH_ph13-self.C_NaOH_2)
       
       #Calculate the total outlet volumetric flow rate from 2nd step L/h
       self.Qtot_out_2=self.Qtot_out_1+self.QNaOH_2_st+self.QNaOH_2_add
      
       #Calculate the outlet mass flow rate of calcium hydroxide produced during the 2nd step kg/hr 
       self.M_CaOH2_2=(self.Qtot_out_1*self.CCa_out_1*(self.conv_2/100)*MW_CaOH)/1000 #kg/hr
      
       #Calculate the outlet mass flow rate of magnesium hydroxide produced during the 2nd step kg/hr 
       self.M_MgOH2_2=self.Qtot_out_1*self.CMg_out_1*MW_MgOH/1000 
       
       #Calculate magma density: the quantity of solids produced per volume of slurry kg/l
       self.magma_d_2=(self.M_MgOH2_2+self.M_CaOH2_2)/self.Qtot_out_2 
      
       #Calculate total outlet volumetric flow rate for 2nd step in L/h
       self.Qout_2=self.Qtot_out_2-self.M_CaOH2_2/d_caoh_2-self.M_MgOH2_2/d_mgoh_2
      
       
       #Calculates the outlet ion concentration in mol/L
       self.CNa_out_2=(self.Qtot_out_1*self.CNa_out_1+(self.QNaOH_2_st+self.QNaOH_2_add)*self.C_NaOH_2)/self.Qtot_out_2       
       self.CCa_out_2=(self.QCa_in_2*(1-self.conv_2/100))/self.Qtot_out_2
       self.CCl_out_2=(self.Qtot_out_1*self.CCl_out_1)/self.Qtot_out_2
       self.CK_out_2=(self.Qtot_out_1*self.CK_out_1)/self.Qtot_out_2
       self.CMg_out_2=0
       self.CSO4_out_2=(self.Qtot_out_1*self.CSO4_out_1)/self.Qtot_out_2
       
       #Calculate ph of solution during the second step 
       self.ph_2=14+math.log10(0.1)
    
       return self.QCa_in_2, self.QNaOH_2_st, self.COH_ph13, self.QNaOH_2_add, self.Qtot_out_2, self.M_CaOH2_2, self.M_MgOH2_2, self.magma_d_2, self.Qout_2, self.CNa_out_2, self.CCa_out_2, self.CCl_out_2, self.CK_out_2, self.CMg_out_2, self.CSO4_out_2,  self.ph_2

class HClAddition:
    def __init__(self, Qout_2, Cout_all_m, MW_cl, ph_2, HCl_conc):
        self.Qout_2 = Qout_2
        self.Cout_all_m = Cout_all_m
        self.MW_cl = MW_cl
        self.ph_2 = ph_2
        self.HCl_conc=HCl_conc

    def calculate_HCl_addition(self, Cout_mfpfr_g):
        OH_initial = math.pow(10, -14) / math.pow(10, -self.ph_2)
        OH_final = math.pow(10, -14) / math.pow(10, -7)
        
        #Calculate the volume of HCl 1M in liters, Qout_f
        QHCl = (self.Qout_2 * OH_initial - self.Qout_2 * OH_final) / (OH_final + self.HCl_conc)  
        self.Qout_f = self.Qout_2+QHCl

        # Calculate the outlet concentration of chloride ions
        C_cl_out = (self.Cout_all_m[1] * self.Qout_2 + QHCl * self.HCl_conc) / self.Qout_f  # mol/l
        self.Cout_all_m[1] = C_cl_out
        C_cl_out = C_cl_out * self.MW_cl  # g/l
        Cout_mfpfr_g[1] = C_cl_out

        for i in range(2, 6):
            Cout_mfpfr_g[i] = Cout_mfpfr_g[i] * self.Qout_2 / self.Qout_f

        Cout_mfpfr_g[0] = Cout_mfpfr_g[0] * self.Qout_2 / self.Qout_f
        # Return the volume of HCl added and the outlet concentration of chloride ions
        return QHCl, Cout_mfpfr_g
     
  # Get the outlet flow rate

#%%Energy consumption 
class energycons:
    def energycalc(Qtot, QNaOH, Qin, QNaOH_1, QNaOH_2_add, QNaOH_2_st, dp, npump):
        Qtot=Qtot
        QNaOH=QNaOH
        Qin=Qin
        QNaOH_2_st=QNaOH_2_st
        QNaOH_2_add=QNaOH_2_add
        QNaOH_1=QNaOH_1        
        
        #Calculate the energy required for pumping in kW
        Epump_1=(Qin*dp+QNaOH_1*dp)*1e5/3600/(1000*npump) #(W)
        Epump_2=((Qin+QNaOH_1)*dp+(QNaOH_2_add+QNaOH_2_st)*dp)*1e5/3600/(1000*npump)       #(W)
        return Epump_1, Epump_2
                    




# Scaling Functions for Technology Processes
"""
    Scale-up function for technology-specific processes.
    
    This function calculates the scaled-up equipment cost or other relevant 
    parameter (C2) based on the equipment cost (C1), the original 
    mass or volume (M1), the scaled mass or volume (M2), and the specific 
    technology being used.
    
    Parameters:
    -----------
    C1 : float
        The original equipment cost or value of the parameter to be scaled.
    
    M1 : float
        The original mass or volume associated with C1.
    
    M2 : float
        The scaled mass or volume for which the new concentration or 
        parameter value (C2) will be calculated.
    
    x : float
        The scaling factor is defined based on the technology/process. 
        For these technologies, the scaling factor x is set to 0.6.
    
    Returns:
    --------
    C2 : float
        The scaled concentration or value of the parameter after applying 
        the scaling factor.
"""

def scaleup(C1,M1,M2):
    x=0.6        
    C2=C1*(M2/M1)**x
    return C2


#costs calculations functions 
from desalsim import constants 
#%%
#symbols:
#hr-> hours
#lf -> lifetime 
#el -> electricity
#s-> steam
#pr -> price 
#c-> cost
#eq-> equipment 
#conc -> consumption 
#wat -> water 
#cw -> cooling water 
#E -> Energy 
#prd -> product mass rate 

#%%constants
class econom:
    def __init__(self, eq_c, el_conc, s_conc, chem1_conc, chem1_pr,chem2_conc, chem2_pr, cw_conc, wat_conc):
        """
        Initialize an instance of the economic costs.

        Args:
            equipment_cost (float): Cost of equipment in euros.
            electricity_consumption (float): Electricity consumption in kWh/year.
            steam_consumption (float): Steam consumption in kWh/year.
            chemical1_concentration (float): Concentration of chemical 1 in solution (e.g., 1M).
            chemical1_price (float): Price of chemical 1 per unit (e.g., euro/L).
            chemical2_concentration (float): Concentration of chemical 2 in solution (e.g., 1M).
            chemical2_price (float): Price of chemical 2 per unit (e.g., euro/L).
            cooling_water_consumption (float): Cooling water consumption in kg/year.
            water_consumption (float): Water consumption in kg/year.
        """
        self.eq_c= eq_c
        self.el_conc=el_conc
        self.s_conc=s_conc
        self.chem1_conc=chem1_conc
        self.chem1_pr=chem1_pr
        self.chem2_conc=chem2_conc
        self.chem2_pr=chem2_pr
        self.cw_conc=cw_conc
        self.wat_conc=wat_conc
        
    def capex_calc(self, capex_assumptions):
        """
        Calculate the capital expenditure (CAPEX) of the unit.
        
        Args:
            inst_c (float): Installation cost in euros 
            buil_c (float): Building, process and auxillary cost in euros
            land_c (float): Land cost in euros
            hard_c (float): Hardware costs are the sum of costs on purchased equipment and installation in euro 
            dir_c (float):  Direct costs in euro. Direct costs consist of purchased equipment, Purchased-equipment installation, Instrumentation and controls, 
                            Piping, Electrical systems, Buildings (including services), Yard improvements,Service facilities, Land 
            ind_c (float):  Indirect costs in euro. Indirect costs consist of engineering and supervision costs, legal expenses, Construction expenses, 
                            Constructor's fee, Contingency
            fix_c (float):  fixed-capital investment in euro. fixed-capital investment is the capital necessary for the in- stalled process equipment 
                            with all components that are needed for complete process operation. Expenses
            work_c (float): working capital in euro 
            t_capital_inv (float): total capital investment 
            
        """
        #Calculate Installation cost in euros
        self.inst_c=capex_assumptions[0]*self.eq_c  
        
        #Calculate Building, process and auxillary cost in euros
        self.buil_c=capex_assumptions[1]*self.eq_c   
        
        #Calculate Land cost in euros
        self.land_c=capex_assumptions[2]*self.eq_c   
        
        #Calculate Hardware costs in euro 
        self.hard_c=self.eq_c+self.inst_c 
        
        #Calculate Direct costs in euro
        self.dir_c= self.hard_c+self.buil_c+ self.land_c 
        
        #Calculate Indirect costs in euro
        self.ind_c=capex_assumptions[3]*self.dir_c 
        
        #Calculate fixed-capital investment in euro
        self.fix_c=self.dir_c+self.ind_c 
        
        #Calculate working capital in euro
        self.work_c=capex_assumptions[4]*self.fix_c
        
        #Calculate total capital investment in euro
        self.t_capital_inv=self.fix_c+self.work_c 
        return self.t_capital_inv
      
        
    
    def opex_calc(self, hr, el_pr, s_pr, cw_pr, w_pr, economic_assumptions):
        """
        Calculate the capital expenditure (OPEX) of the unit.
        
        Args:
            E_el (float): electricity cost in euro 
            E_th (float): thermal energy cost in euro 
            t_E_c total(float): energy costs in euro 
            chem_c (float): cost for chemicals in euro 
            cw_c (float): cost for cooling water in euro 
            wat_c (float): cost for water consumption in euro 
            main_c (float): maintenance cost  in euro 
            oper_sup_c (float): operating suppliers cost  in euro 
            oper_lab_c (float): operating labor in euro 
            super_c (float): direct supervisory and clerical labor in euro 
            lab_c (float): Laboratory charges in euro 
            pat_c (float): Patents and royalties in euro 
            fix_char (float): Fixed charges in euro 
            over_c (float): plant overhead costs in euro 
            OPEX (float): Operating costs in euro. It consists of utilities, maintenance, operating supplies, operating labor, direct supervisory and clerical labor, 
                         laboratory charges, patents and royalties, fixed charges, and plant overhead cost             
         """
         
        """utilities cost calculation """        


        #Calculate Electricity cost in euro 
        self.E_el=self.el_conc*hr*el_pr 
        
        #Calculate Thermal energy cost in euro 
        self.E_th=self.s_conc*hr*s_pr  
        
        #Calculate Total energy cost in euro 
        self.t_E_c=self.E_el+self.E_th     
        
        #Calculate cost for chemicals in euro 
        self.chem_c=self.chem1_conc*hr*self.chem1_pr+self.chem2_conc*hr*self.chem2_pr  
        
        #Calculate cost for cooling water in euro 
        self.cw_c=self.cw_conc*hr*cw_pr 
        
        #Calculate cost for water in euro 
        self.wat_c=self.wat_conc*hr*w_pr  
        
        """Operating costs calculation """
        main_c_percent, oper_sup_c_percent, oper_lab_c_percent, super_c_percent, lab_c_percent, pat_c_percent, fix_char_percent, over_c_percent, norm_factor = economic_assumptions
        
        #Calculate maintenance cost in euro
        self.main_c=main_c_percent*self.fix_c        
        
        #Calculate operating suppliers cost  in euro 
        self.oper_sup_c=oper_sup_c_percent*self.main_c  
        
        #Calculate operating labor in euro
        self.oper_lab_c=(self.t_E_c+self.chem_c+self.wat_c+self.cw_c+self.main_c+self.oper_sup_c)/norm_factor*oper_lab_c_percent     
        
        
        #Calculate direct supervisory and clerical labor in euro 
        self.super_c=super_c_percent*self.oper_lab_c 
        
        #Calculate Laboratory charges in euro 
        self.lab_c=lab_c_percent*self.oper_lab_c
        
        #Calculate Patents and royalties in euro
        self.pat_c=(self.t_E_c+self.chem_c+self.wat_c+self.cw_c+self.main_c+self.oper_sup_c)/norm_factor*pat_c_percent 
        
        #Calculate Fixed charges in euro
        self.fix_char=(self.t_E_c+self.chem_c+self.wat_c+self.cw_c+self.main_c+self.oper_sup_c)/norm_factor*fix_char_percent  
        
        #Calculate plant overhead costs in euro
        self.over_c=(self.t_E_c+self.chem_c+self.wat_c+self.cw_c+self.main_c+self.oper_sup_c)/norm_factor*over_c_percent
        
        #Calculate OPEX in euro 
        self.opex=self.t_E_c+self.chem_c+self.wat_c+self.cw_c+self.main_c+self.oper_sup_c+self.oper_lab_c+self.super_c+self.lab_c+self.pat_c+self.fix_char+self.over_c
        
        return self.opex
#%%
class revenue:
    """
    Calculate the revenues from selling products  of the unit.
    
    Args:
        prd_name (str): name of product 
        rev_prd (float): reveneues from product i in euro  
    """
    def __init__(self, prd, prd_name):
        self.prd=prd
        self.prd_name=prd_name
        
        
    def rev(self, hr, w_pr, nacl_pr, mgoh2_pr,na2so4_pr, naoh_pr, hcl_pr ):
        #density
        d_naoh=1.04 #kg/l for 1M solution 
        d_hcl=1.0585#kg/l for 1M solution 
        
        if self.prd_name=="Water":
            self.rev_prd=self.prd*w_pr*hr #euro/year
        elif self.prd_name=="NaCl":
            self.rev_prd=self.prd*nacl_pr*hr #euro/year
        elif self.prd_name=="Mg(OH)2":
            self.rev_prd=self.prd*mgoh2_pr*hr #euro/year
        elif self.prd_name=="Na2SO4":
            self.rev_prd=self.prd*na2so4_pr*hr #euro/year
        elif self.prd_name=="NaOH":
            self.prd=self.prd*d_naoh
            self.rev_prd=self.prd*naoh_pr*hr #euro/year
        elif self.prd_name=="HCl":
            self.prd=self.prd*d_hcl
            self.rev_prd=self.prd*hcl_pr*hr #euro/year




#setup.py

from setuptools import setup, find_packages

with open("README.md", "r") as f: 
    description =f.read()

setup(
      name='desalsim',
      version='1.0.9.1',
      packages=find_packages(),
      url="https://github.com/rodoulak/desalsim.git",
      author="rodoulak",
      author_email="r.ktori@tudelft.nl",
      license="MIT",
      install_requires=[
          "numpy", #==1.26.4
          "pandas",  # Include pandas as a dependency ==2.2.2
          "scipy", # For figures ==1.13.0
          "matplotlib", # For figures ==3.8.4
          "scikit-learn", #==1.4.2
          "plotly", #==5.22.0
          "openpyxl",  #read data from excel file ==3.1.2
      ],
      long_description= description,
      long_description_content_type="text/markdown",
      )



#scale up faction 
#x=0.8 for membrane technologies 
#x=0.6 for volume increase technologies  

def scaleup_eq(C1,M1,M2,tec):
    if tec=="NF" or tec=="EDBM" or tec=="MED":
        x=0.6  
    else:
        x=0.6        
    C2=C1*(M2/M1)**x
    return C2

def scaleup(C1,M1,M2):
    x=0.6        
    C2=C1*(M2/M1)**x
    return C2


import unittest
from desalsim.nanofiltration_unit_f import NFMass, OsmoticPressure, NfEnergy

class TestNanofiltrationUnit(unittest.TestCase):

    def test_nfmass(self):
        # Test NFMass calculation
        components = ['Na', 'Cl', 'K', 'Mg', 'Ca', 'SO4']
        self.C_in = [12.33, 21.67, 0.45, 1.39, 0.45, 3.28]
        rjr_values = [0.16, 0.29, 0.21, 0.98, 0.95, 0.98] #Ions rejection rates based on membrane characteristics (units: -)
        Wrec = 0.7 # Water recovery based on membrane characteristics (units: -)
        Qf=1000
        
        nf_mass = NFMass(components[0], self.C_in[0], rjr_values[0], Wrec, Qf)  # Example values
        self.assertAlmostEqual(nf_mass.Qperm, 700, delta=10)  # Expected Qperm: 875 kg/hr
        self.assertAlmostEqual(nf_mass.Qconc, 300, delta=10)  # Expected Qconc: 2625 kg/hr
        # Add more tests for other attributes if needed

    def test_osmotic_pressure(self):
        # Test OsmoticPressure calculation
        self.C_in = [12.33, 21.67, 0.45, 1.39, 0.45, 3.28]
        self.z_values = [1, -1, 1, 2, 2, -2]
        osmotic_pressure = OsmoticPressure(self.C_in, self.z_values, 293)  # Example values
        self.assertAlmostEqual(osmotic_pressure.osmotic_pressure_calculation(), 32.86, delta=1)  # Expected osmotic pressure: 22.56 bar

    def test_nfenergy(self):
        # Test NfEnergy calculation
        nf_energy = NfEnergy(57.8, 32.86, 22.9, 2, 1018.1, 700.0, 1000,1028.28, 0.8)  # Example values
        result = nf_energy.calculate_energy_consumption()
        # Add assertions to test the result dictionary
        self.assertAlmostEqual(result['Applied pressure (Bar)'], 24.4, delta=2)  # Expected applied pressure: 12.56 bar


if __name__ == '__main__':
    unittest.main()



import numpy as np
from desalsim import scaleup
from desalsim import constants 
import math
from desalsim.density_calc import density_calc 

#%% Calculations 
"""
A class used to represent Mass Balance for Electrodialysis

...

Attributes
----------
    MWs : float
        Molecular weight of NaCl (g/mol)
    MWw : float
        Molecular weight of water (g/mol)
    R : float
        Resistance of rinse stream (ohm)
    Rp : float
        Resistance of polarization (ohm)
    A : float
        Active area of cell-pair (m^2)
    F : float
        Faraday constant (C/mol)
    T : float
        Temperature in Kelvin
    dp : float
        Parameter dp
    npump : float
        Pump efficiency
    rho_w : float
        Density of water (kg/m^3)
    D : float
        Diffusion coefficient (m^2/s)
    tcu : float
        tcu parameter
    veloc : float
        Velocity (m^2/s)
    h : float
        Height (mm)
    Sh : float
        Sh parameter
    Mem_eff : float
        Membrane efficiency
    Ncp : integer
        Number of cell-pairs
    Qed_in : float
        Inlet flow rate (l/h)
    Qed_in_c : float
        Concentrate inlet flow rate (l/h)
    Qed_in_d : float
        Dilute inlet flow rate (l/h)


Methods
-------
    Ts_cp(S): 
        Transport number for salt in concentrate compartment
    Tw_cp(Sc, Sd): 
        Transport number for water in concentrate compartment
    Ls_cp(Sc, Sd): 
        Permeability for salt in concentrate compartment
    Lw_cp(S): 
        Permeability for water in concentrate compartment
    p_osmo(S): 
        Osmotic pressure in concentrate compartment
    p_osmo2(S): 
        Alternative osmotic pressure calculation
    dC(Ts_cp): 
        Change in concentration
"""
class ElectrodialysisCalc:
    def __init__(self):
        pass

    def Ts_cp(S):
        """Transport number for salt in concentrate compartment"""
        Ts_cp=-4e-6*S**2 + 4e-5*S + 0.96
        return Ts_cp
    
    def Tw_cp(Sc,Sd):
        """Transport number for water in concentrate compartment"""
        return -4e-5*Sc**2 - 1.9e-2*Sd + 11.2
    
    def Ls_cp(Sc, Sd):
        """Permeability for salt in concentrate compartment"""
        return min(2e-12*Sd**2 - 3e-10*Sd + 6e-8, 2e-12*Sc**2 - 3e-10*Sc + 6e-8)
    
    def Lw_cp(S):
        """Permeability for water in concentrate compartment"""
        return 5*S**(-0.416)
    
    
    def p_osmo(S, T, MWs): 
        """calculation for Osmotic pressure in concentrate compartment"""
        C1=S/MWs*constants.MW_Na/constants.MW_Na
        C2=S/MWs*constants.MW_cl/constants.MW_cl
        sum_Ci=sum([C1, C2])
        p_osmo=0.0831446261815324*sum_Ci*T
        return p_osmo
    
    def dC(Ts_cp, tcu, D, Ij,  h, Sh ):
        """Change in concentration"""
        F = 96485.3329  # Faraday constant (C/mol)
        Tcu=(Ts_cp+1)/2
        dC=-(Tcu-tcu)/D*(Ij/F)*(2*h/1000/Sh)
        return dC




#Crystal Size Disribution in a Eutectic System

'''Model for the estimation of the distribution of crystal sizes 
during the cooling and eutectic freeze crystallization of an
aqueous solution of Compound 1 (Na2SO4/Sodium sulfate), that initially includes traces of Compound 2 (considered as impurity: Extracellular Polymeric Substances/Sodium chloride), in a batch/continuous reactor'''

#%%
#constants 
dp=0.1
dp_slurry=1
npump=0.8

#Import the libraries that will be used
import math
import pandas as pd
pi=math.pi
exp=math.exp
log=math.log

import matplotlib.pyplot as plt
fig1=plt.figure(1)
fig2=plt.figure(2)
fig3=plt.figure(3)
fig4=plt.figure(4)
fig5=plt.figure(5)
fig6=plt.figure(6)

import time
import numpy as np
from scipy import interpolate
import scipy.interpolate as interpolate
from sklearn.linear_model import LinearRegression
from collections import namedtuple
from desalsim import scaleup
from desalsim.density_calc import density_calc 
# import pitzer_model_3phases_w
from desalsim import constants 
#Set initial time to calculate the elapsed time of the calculations
time0=time.time()
MW_so4=32.064+4*15.999

#input data
Qf=100 #m3/hr
C_i_in=[20.16, 27.87, 0.50, 0.0, 0.12, 14.66]

T_initial =273+25
d_in=density_calc(25, sum(C_i_in))
#%%

#Declaration of constant process parameters - Part 1: Process Parameters

Vreactor=Qf/0.7            #Volume of the reactor : m^3

#Influx
Fin_water=0                                                                                                         #Influx of water (in mother liquor) in the reactor : kg/s
Fin_compound1=0                                                                                                     #Influx of Compound 1 (in mother liquor) in the reactor : kg/s
Fin_compound2=0                                                                                                     #Influx of Compound 2 (in mother liquor) in the reactor : kg/s
Fin_mother_liquor=Fin_water+Fin_compound1+Fin_compound2                                                             #Influx of mother liquor in the reactor : kg/s
Fin_liquid_tot=Fin_mother_liquor                                                                                    #Total influx of liquid in the reactor : kg/s

Fin_ice=0                                                                                                           #Influx of ice crystals in the reactor : kg/s
Fin_compound1_cr=0                                                                                                  #Influx of Compound 1 crystals in the reactor : kg/s
Fin_solids_tot=Fin_ice+Fin_compound1_cr                                                                             #Total influx of solid crystals in the reactor : kg/s

#Outflux
Fout_water=0                                                                                                        #Outflux of water (in mother liquor) from the reactor : kg/s
Fout_compound1=0                                                                                                    #Outflux of Compound 1 (in mother liquor) from the reactor : kg/s
Fout_compound2=0                                                                                                    #Outflux of Compound 2 (in mother liquor) from the reactor : kg/s
Fout_mother_liquor=Fin_water+Fin_compound1+Fin_compound2                                                            #Outflux of mother liquor from the reactor : kg/s
Fout_liquid_tot=Fout_mother_liquor                                                                                  #Total outflux of liquid from the reactor : kg/s

Fout_ice=0                                                                                                          #Outflux of ice crystals from the reactor : kg/s
Fout_compound1_cr=0                                                                                                 #Outflux of Compound 1 crystals from the reactor : kg/s
Fout_solids_tot=Fout_ice+Fout_compound1_cr                                                                          #Total outflux of solid crystals from the reactor : kg/s

Fout_top=0                                                                                                          #Outflux from the top (ice and mother liquor) : kg/s
Fout_bottom=0                                                                                                       #Outflux from the bottom (Compound 1 and mother liquor) : kg/s

#Net Heat flux 
Cooling_rate=2500000                                                                                                 #Assumed NET cooling rate to cool down : W = J/s :Assumption !!! Includes the heat removed by the heat exchanger and the heat provided by the environment and the system
    
#General Parameters
classes=300                                                                                                         #Number of crystal size classes
maxl=1e-3                                                                                                           #Maximum crystal length : m
deltal=maxl/classes                                                                                                 #Length interval per class : m
deltat=5                                                                                                           #Time interval : sec  

class_length=np.array([])                                                                                           #Array of the lengths for the crystal size classes used for plotting
for i in range(0,classes):
       class_length=np.append(class_length,((i*deltal)+deltal))
       
#Storage lists for different values
t_store=np.array([])
sigma_compound1_store=np.array([])
DT_ice_store=np.array([])
M_ice_store=np.array([])
M_compound1_cr_store=np.array([])
Treactor_store=np.array([])


#%%

#Declaration of constant process parameters - Part 2: WATER Properties library

#Physical properties
d_water=1*1000                                                                                                      #Density of water (25oC) : kg/m^3 (source:Wikipedia) 
Cp_water=4.1813*1000                                                                                                #Heat capacity of water (25 oC) : J/(kg * K) (source:Wikipedia)
#%%

#Declaration of constant process parameters - Part 3: ICE Properties 

#Kinetic and Shape Parameters

#Ice
kg_ice=5e-8                                                                                                         #Growth rate constant for ice crystals : m/(K*s) (source: Characterization and Population....)
kb_ice=5e4                                                                                                          #Nucleation rate constant for ice crystals : 1/(m^2*s*K^2) (source: Characterization and Population....)
kd_ice=1e-10                                                                                                        #ASSUMED Dissolution rate constant for ice crystals : m/(K*s)

r_ice=10                                                                                                            #L/h aspect ratio of ice crystals, assuming a disk shape where length=10*height : -
kv_ice=pi/(4*r_ice)                                                                                                 #Shape factor of ice crystals, where V=kv*l^3 (assuming disk shape and l=longest side) : -
ka_ice=pi*(1/r_ice+1)                                                                                               #Surface factor of ice crystals, where A=ka*l^2 (assuming disk shape and l=longest side) : -

#Physical Properties
d_ice=0.917*1000                                                                                                    #Density of ice (25oC) : kg/m^3 (source:https://hypertextbook.com/facts/2000/AlexDallas.shtml)
Cp_ice=2.05*1000                                                                                                    #Heat capacity of ice (0 oC) : J/(kg * K) (source:https://www.engineeringtoolbox.com/ice-thermal-properties-d_576.html)

#Thermodynamic Properties
DHfus_ice=333.55*1000                                                                                               #Enthalpy of fusion of ice : J/kg (source:Wikipedia)
#%%
#Declaration of constant process parameters - Part 5: Sodium Sulfate Properties 

#Physical Properties
MW_Na2SO4=142.04                
MW_Na2SO4_10H2O=322.192258                                                                         #Molecular weight of sodium sulfate : g/mol
d_Na2SO4=2.66*1000                                                                                                  #Density of sodium sulfate (25oC) : kg/m^3 (source:???) 
Cp_Na2SO4=(128.2/MW_Na2SO4)*1000                                                                                      #Heat capacity of sodium sulfate (25 oC) : J/(kg * K) (source:source: https://en.wikipedia.org/wiki/Sodium_sulfate_(data_page))

#Kinetic and Shape Parameters
kg_Na2SO4=(43960*exp(-24600/((8.314)*T_initial )))/(1000*3600)                                                             #Growth rate constant for sodium sulfate crystals : m/s : Mirabilite at 25oC(source: crystallization kinetics of sodium sulfate... #UBC TBC)
kb_Na2SO4=(0.064*exp(70900/((8.314)*T_initial )))*1000/(3600*d_Na2SO4)                                                     #Nucleation rate constant for sodium sulfate crystals : 1/(kg*s) : Mirabilite (source: crystallization kinetics of sodium sulfate... #UBC TBC)
kd_Na2SO4=1e-10                                                                                                     #ASSUMED Dissolution rate constant for sodium sulfate crystals : m/s

r_Na2SO4=1/10                                                                                                       #L/h aspect ratio of sodium sulfate crystals, assuming a cylindrical shape where length=height/10 : -
kv_Na2SO4=(pi*r_Na2SO4**2)/4                                                                                        #Shape factor of sodium sulfate crystals, where V=kv*h^3 (assuming cylindrical shape and h=longest side) : -
ka_Na2SO4=(pi*r_Na2SO4)*(1+r_Na2SO4/2)                                                                              #Surface factor of sodium sulfate crystals, where A=ka*h^2 (assuming cylindrical shape and h=longest side) : -

#Thermodynamic Properties
DHfus_Na2SO4=252*1000                                                                                               #Enthalpy of fusion of sodium sulfate : J/kg (source:Solubility of Na2SO4...)

#Create the library with the constant parameters for sodium sulfate
Na2SO4Parameters = namedtuple('Na2SO4Parameters', ['MW','d','Cp',
                                                   'DHfus','kg','kb','kd','r',
                                                   'kv','ka'])

Na2SO4Par = Na2SO4Parameters(MW=MW_Na2SO4,        
                             d=d_Na2SO4,        
                             Cp=Cp_Na2SO4,    
                             DHfus=DHfus_Na2SO4,  
                             kg=kg_Na2SO4,      
                             kb=kb_Na2SO4,        
                             kd=kd_Na2SO4,
                             r=r_Na2SO4,
                             kv=kv_Na2SO4,         
                             ka=ka_Na2SO4)  
#%%
#Declaration of constant process parameters - Part 7: Sodium Chloride Properties

#Physical Properties
MW_NaCl=58.4                                                                                                        #Molecular weight of NaCl : g/mol (ASSUMED??????) 
d_NaCl=2.16*1000                                                                                                    #Density of NaCl (25oC) : kg/m^3 (source:https://pubchem.ncbi.nlm.nih.gov/compound/sodium_chloride#section=Solubility) 
Cp_NaCl=(50.5/MW_NaCl)*1000                                                                                         #Heat capacity of NaCl (25 oC) : J/(kg * K) (source:http://webbook.nist.gov/cgi/cbook.cgi?ID=C7647145&Type=JANAFS&Table=on)

#Create the library with the constant parameters for sodium chloride
NaClParameters = namedtuple('NaClParameters', ['MW','d','Cp'])

NaClPar = NaClParameters(MW=MW_NaCl,        
                       d=d_NaCl,        
                       Cp=Cp_NaCl)
#%%

#Assign all the used (non-constant) variables to a class
 
class Variables:
    def __init__(self,t,N_ice,N_compound1_cr,V_ice,V_compound1_cr,V_solids_tot,M_ice,M_compound1_cr,M_solids_tot,Ntot_ice,Ntot_compound1_cr,Ntot,C_compound1,C_compound2,Treactor,V_liquid_tot,V_mother_liquor,V_water,V_compound1,V_compound2,M_water,M_compound1,M_compound2,M_mother_liquor,M_liquid_tot,Vtot,Mtot,solids_weight_percentage,V_ice_distr,V_compound1_cr_distr,V_ice_previous_loop,V_compound1_cr_previous_loop):
        self.t=t
        self.N_ice=N_ice
        self.N_compound1_cr=N_compound1_cr
        self.V_ice=V_ice
        self.V_compound1_cr=V_compound1_cr
        self.V_solids_tot=V_solids_tot
        self.M_ice=M_ice
        self.M_compound1_cr=M_compound1_cr
        self.M_solids_tot=M_solids_tot
        self.Ntot_ice=Ntot_ice
        self.Ntot_compound1_cr=Ntot_compound1_cr
        self.Ntot=Ntot
        self.C_compound1=C_compound1
        self.C_compound2=C_compound2
        self.Treactor=Treactor
        self.V_liquid_tot=V_liquid_tot
        self.V_mother_liquor=V_mother_liquor
        self.V_water=V_water
        self.V_compound1=V_compound1
        self.V_compound2=V_compound2
        self.M_water=M_water
        self.M_liquid_tot=M_liquid_tot
        self.M_mother_liquor=M_mother_liquor
        self.M_compound1=M_compound1
        self.M_compound2=M_compound2
        self.Mtot=Mtot
        self.Vtot=Vtot
        self.solids_weight_percentage=solids_weight_percentage
        self.V_ice_distr=V_ice_distr
        self.V_compound1_cr_distr=V_compound1_cr_distr
        self.V_ice_previous_loop=V_ice_previous_loop
        self.V_compound1_cr_previous_loop=V_compound1_cr_previous_loop
       
Var=Variables

#%%

#Function for the selection of the compounds that will be used for the simulation. 

def Compounds():
    global Compound1,Compound2,tck,lm,Linear_regr,C_compound1_initial,C_compound2_initial
    
    #Initial concentrations of the different compounds
    #Initial sodium chloride concentration in the mother liquor in the reactor : molal
    C_so4_in=(C_i_in[5]/constants.MW_so4)/(d_in/1000) #mol/kg
    print("C_so4_i is "+str(C_so4_in))
    C_nacl_in=(C_i_in[0]/constants.MW_Na-2*C_so4_in)/(d_in/1000) #mol/kg
    print("C_nacl_in" + str(C_nacl_in))
    C_Na2SO4_initial=np.array([C_so4_in]).reshape(-1,1)   #np.array([0.364]).reshape(-1,1)                                                                                             #Initial sodium sulfate concentration in the mother liquor in the reactor : molal
    C_NaCl_initial=np.array([C_nacl_in]).reshape(-1,1)                               # 0.001                                                        
    #The user is given a choice for Compound 1 (which is the main compound crystallizing), in this case sodium sulfate  
    
    Compound1=Na2SO4Par
       
    print('Compound 1 : Sodium Sulfate')
        
    #Initial concentration of Compound 1
    C_compound1_initial=C_Na2SO4_initial
    print("C_compound1_initial is "+str(C_compound1_initial))
    T_Na2SO4_sol_line=[]
    C_Na2SO4_sol_line=[]
    #Interpolation/Extrapolation of the Na2SO4 solubility line with a cubic spline 
    import os
    import pandas as pd

    # Get the directory where the script is located
    current_directory = os.path.dirname(__file__)

    #   Define the filename 
    filename = "efc_results.xlsx"

    # Combine the directory path with the filename
    file_path = os.path.join(current_directory, filename)

    # Read the Excel file
    df = pd.read_excel(file_path, sheet_name="C_Na2SO4_sol_line")

    #Sodium sulfate (Mirabilite) Solubility Line, Data from source: The solution ref on dropbox (1100-1200; P.1120)
    T_Na2SO4_sol_line= df.iloc[:, 1].tolist()
    C_Na2SO4_sol_line=df.iloc[:, 0].tolist()
    C_Na2SO4_sol_line=np.array(C_Na2SO4_sol_line)

    print("T_Na2SO4_sol_line "+str(T_Na2SO4_sol_line))

    print("C_Na2SO4_sol_line "+str(C_Na2SO4_sol_line))
                                                            

    tck= interpolate.splrep(T_Na2SO4_sol_line, C_Na2SO4_sol_line, k=1)
    print("tck is " +str(tck))

    #Interpolation/Extrapolation of the ice line using linear regression
    df2 = pd.read_excel(file_path, sheet_name="ICE_sol_line")

    Fr_point_depr_Na2SO4= np.array(df2.iloc[:, 1].tolist())
    C_Na2SO4_ice_line=df2.iloc[:, 0].tolist()    
    print("C_Na2SO4_ice_line is "+str(C_Na2SO4_ice_line))
    #Ice line in the Na2SO4-Water phase diagram : Source : CONCENTRATIVE PROPERTIES OF AQUEOUS SOLUTIONS:DENSITY, REFRACTIVE INDEX, FREEZING POINT DEPRESSION, AND VISCOSITY

    T_fr_Na2SO4=273+Fr_point_depr_Na2SO4    

    print("T_fr_na2so4 is " + str(T_fr_Na2SO4))
    lm = LinearRegression()
    Linear_regr=lm.fit(np.array(C_Na2SO4_ice_line).reshape(-1, 1), T_fr_Na2SO4)
    #The user is also given a choice for Compound 2 (initially considered as impurity), in this case sodium chloride      
    Compound2=NaClPar

    print('Compound 2 : Sodium Chloride')
        
    #Initial concentration of Compound 2
    C_compound2_initial=C_NaCl_initial
    return(C_Na2SO4_sol_line, tck)
   
#%%

#Function for the initialization of all the (non-constant) variables, that are assigned to the Variables class 
def Initialize(Var):
    global Compound1,Compound2,C_compound1_initial,C_compound2_initial
    #Initial values
    
    Var.t=0                                                                                                             #Time : sec
    
    Var.N_ice=np.array([])                                                                                              #Number of ice crystals in each class in the reactor : absolute number of ice crystals
    Var.N_compound1_cr=np.array([])                                                                                     #Number of compound 1 crystals in each class in the reactor : absolute number of Na2SO4 crystals
    
    Var.V_ice=0                                                                                                         #Total volume of ice crystals in the reactor: m^3
    Var.V_compound1_cr=0                                                                                                #Total volume of compound 1 crystals in the reactor: m^3
    Var.V_solids_tot=Var.V_ice+Var.V_compound1_cr                                                                       #Total volume of solids in the reactor : m^3
    
    Var.V_ice_distr=np.zeros(classes)                                                                                   #Volume distribution of ice crystals in each class : m^3
    Var.V_compound1_cr_distr=np.zeros(classes)                                                                          #Volume distribution of compound 1 crystals in each class : m^3
    
    Var.M_ice=0                                                                                                         #Total mass of ice in the reactor : kg
    Var.M_compound1_cr=0                                                                                                #Total mass of compound 1 crystals in the reactor : kg
    Var.M_solids_tot=Var.M_ice+Var.M_compound1_cr                                                                       #Total mass of solids in the reactor : kg
    
    Var.Ntot_ice=0                                                                                                      #Total number of ice crystals in the reactor
    Var.Ntot_compound1_cr=0                                                                                             #Total number of compound 1 crystals in the reactor
    Var.Ntot=Var.Ntot_ice+Var.Ntot_compound1_cr                                                                         #Total number of crystals in the reactor
    
    Var.C_compound1=C_compound1_initial                                                                                 #Initial compound 1 concentration in the mother liquor in the reactor : molal = mol compound 1/kg water
    Var.C_compound2=C_compound2_initial                                                                                 #Initial compound 2 concentration in the mother liquor in the reactor : molal = mol compound 2/kg water
    
    Var.Treactor=T_initial                                                                                                  #Initial temperature of the bulk liquid in the crystallizer : K
    
    Var.V_liquid_tot=0.7*Vreactor                                                                                       #Total volume of liquid in the reactor : m^3 (ASSSUMED to be 70 % of the reactor volume)
    Var.V_mother_liquor=Var.V_liquid_tot                                                                                #Volume of the mother liquor in the reactor = Vwater+V_Na2SO4+Vimp: m^3
    print("v mother liquid is "+str(Var.V_mother_liquor))
    
    Var.V_water=Var.V_mother_liquor/(1+Var.C_compound1*Compound1.MW*d_water/(1000*Compound1.d)+Var.C_compound2*Compound2.MW*d_water/(1000*Compound2.d))       #Volume of water in the reactor : m^3
    Var.V_compound1=(Var.C_compound1*Compound1.MW*Var.V_water*d_water)/(1000*Compound1.d)                                                                     #Volume of compound 1 in the reactor : m^3
    Var.V_compound2=Var.V_mother_liquor-Var.V_water-Var.V_compound1  
    print("Var.V_water is "+str(Var.V_water))

    Var.M_water=Var.V_water*d_water                                                                                     #Mass of water in the reactor : kg
    Var.M_compound1=Var.V_compound1*Compound1.d                                                                         #Mass of compound 1 in the reactor : kg
    print("Var.M_compound1 is "+str(Var.M_compound1))
    Var.M_compound2=Var.V_compound2*Compound2.d                                                                         #Mass of compound 2 in the reactor : kg
    Var.M_mother_liquor=Var.M_water+Var.M_compound1+Var.M_compound2                                                     #Mass of the mother liquor in the reactor :kg
    Var.M_liquid_tot=Var.M_mother_liquor                                                                                #Total mass of liquid in the reactor :kg
    
    Var.Vtot=Var.V_solids_tot+Var.V_liquid_tot                                                                          #Total volume of solids and liquid in the reactor : m^3
    Var.Mtot=Var.M_solids_tot+Var.M_liquid_tot                                                                          #Total mass of solids and liquid in the reactor : kg
    print("Mtot in loop is "+str(Var.Mtot))
    Var.V_ice_previous_loop=Var.V_ice                                                                                   #Total volume of ice crystals stored from the previous time interval : m^3
    Var.V_compound1_cr_previous_loop=Var.V_compound1_cr                                                                 #Total volume of compound 1 crystals stored from the previous time interval : m^3
    
    Var.solids_weight_percentage=(Var.M_solids_tot/Var.Mtot)*100                                                        #Solids weight percentage in the reactor : solids wt % = (total mass of solids in the reactor / total mass in the reactor)*100 % (neglecting the potential solid impurities)
    
    for i in range(0,classes):                                                                                          #Loop over the number of (classes-1) to initialize all the values (The range of the loop is from 0 to classes due to python syntax, the loop actually runs from 0 to classes-1=299 )
        Init_size_distr=2e16*deltal*(exp((-i*deltal)/2.5e-5))                                                           #Initialize the number of crystals in each with this standard distribution (source:???)
        
        Var.N_ice=np.append(Var.N_ice,Init_size_distr)
        Var.N_compound1_cr=np.append(Var.N_compound1_cr,Init_size_distr)
        
        Var.V_ice=Var.V_ice+Var.N_ice[i]*kv_ice*exp(3*log(deltal*(i+1)))
        Var.V_compound1_cr=Var.V_compound1_cr+Var.N_compound1_cr[i]*Compound1.kv*exp(3*log(deltal*(i+1)))
    
    Var.V_ice_distr[0]=Var.N_ice[0]*kv_ice*0.25*(deltal**4)/deltal
    Var.V_compound1_cr_distr[0]=Var.N_compound1_cr[0]*Compound1.kv*0.25*(deltal**4)/deltal
    
    for i in range (1,classes-1):
        
        Var.V_ice_distr[i]=Var.N_ice[i]*kv_ice*0.25*((((i+1)*deltal)**4)-((i*deltal)**4))/deltal
        Var.V_compound1_cr_distr[i]=Var.N_compound1_cr[i]*Compound1.kv*0.25*((((i+1)*deltal)**4)-((i*deltal)**4))/deltal
            
    Var.V_ice_distr[classes-1]=Var.N_ice[classes-1]*kv_ice*0.25*(((classes*deltal)**4)-(((classes-1)*deltal)**4))/deltal
    Var.V_compound1_cr_distr[classes-1]=Var.N_compound1_cr[classes-1]*Compound1.kv*0.25*(((classes*deltal)**4)-(((classes-1)*deltal)**4))/deltal
    
    Var.V_solids_tot=Var.V_ice+Var.V_compound1_cr
        
    Var.M_ice=Var.M_ice+Var.V_ice*d_ice
    Var.M_compound1_cr=Var.M_compound1_cr+Var.V_compound1_cr*Compound1.d

    Var.M_solids_tot=Var.M_ice+Var.M_compound1_cr 
       
    Var.Ntot_ice=Var.Ntot_ice+sum(Var.N_ice)
    Var.Ntot_compound1_cr=Var.Ntot_compound1_cr+sum(Var.N_compound1_cr)
    Var.Ntot=Var.Ntot_ice+Var.Ntot_compound1_cr
    
    Var.Vtot=Var.V_solids_tot+Var.V_liquid_tot
    Var.Mtot=Var.M_solids_tot+Var.M_liquid_tot
    
    Var.V_ice_previous_loop=Var.V_ice                                                               
    Var.V_compound1_cr_previous_loop=Var.V_compound1_cr                                                       
    
    Var.solids_weight_percentage=(Var.M_solids_tot/Var.Mtot)*100                                                                   
    
    return (Var)
#%%
    
 #Function for the estimation of crystal size distribution - Part 1 : Initialization of the local parameters and time interval 
    
def CrystalSizeDistribution():
    global sigma_compound1,DT_ice,V_ice_diff,V_compound1_cr_diff,M_ice_diff,M_compound1_cr_diff,M_ice_reactions,M_compound1_cr_reactions,Compound1,Compound2,tck,lm,Linear_regr        
        
    #Time interval
    Var.t=Var.t+deltat
    
    atot_ice=0                                                                                                      #Total surface area of ice crystals in the reactor in every loop: m^2
    atot_compound1_cr=0                                                                                             #Total surface area of compound 1 crystals in the reactor in every loop: m^2
    
    #Volume and mass difference(increase/decrease) of crystals in every time interval
    V_ice_diff=0                                                                                                    #Volume difference of ice crystals in every time interval : m^3
    V_compound1_cr_diff=0                                                                                           #Volume difference of compound 1 crystals in every time interval : m^3
    
    M_ice_diff=0                                                                                                    #Mass difference of ice crystals in every time interval : kg
    M_compound1_cr_diff=0                                                                                           #Mass difference of compound 1 crystals in every time interval : kg
    
    #Mass of crystals produced/consumed via reacting : growth and/or nucleation and/or dissolution in every time interval
    M_ice_reactions=0                                                                                               #Mass of ice crystals produced/consumed via reacting : growth and/or nucleation and/or dissolution in every time interval : kg
    M_compound1_cr_reactions=0                                                                                      #Mass of compound 1 crystals produced/consumed via reacting : growth and/or nucleation and/or dissolution in every time interval : kg
    
    #Number of crystals produced by nucleation in every time interval
    Nb_ice=0
    Nb_compound1_cr=0
    
    #Number of crystals that  grow in OR out of every class size in every time interval (ASSUMED POSITIVE for growth INTO a class and NEGATIVE for growth OUT of a class)
    Ng_ice=0
    Ng_compound1_cr=0   

    #Number of crystals dissolved in OR out of every class size in every time interval (ASSUMED POSITIVE for dissolution INTO a class and NEGATIVE for dissolution OUT of a class)
    Nd_ice=0
    Nd_compound1_cr=0
    
    #Calculation of the equilibrium values using interpolation/extrapolation of the solubility and ice lines
    C_compound1_eq=interpolate.splev(Var.Treactor, tck, der=0)
    
    C_compound1_eq=float(C_compound1_eq)
    print("C_compound1_eq is "+str(C_compound1_eq))
          
    T_ice_eq_array=lm.predict(Var.C_compound1)
    T_ice_eq=T_ice_eq_array[0]
    print("T_ice_eq is "+str(T_ice_eq))
    
    #Compound 1 Supersaturation  
    sigma_compound1=(Var.C_compound1/C_compound1_eq)-1
    
    #Ice undercooling
    DT_ice=T_ice_eq-Var.Treactor
   
    #%%
    
    #Function for the estimation of crystal size distribution - Part 2 : Compound 1 crystal size distribution / Mass balance
    if sigma_compound1<0:                                                                                             #If there is no supersaturation,the system is undersaturated and there is only dissolution of compound 1 crystals
        #Compound 1 crystals growth rate : m/s
        G_compound1_cr=0
        
        #Compound 1 crystals nucleation rate : crystals/s 
        B_compound1_cr=0
        
        #Compound 1 crystals dissolution rate : m/s
        D_compound1_cr=Compound1.kd*(-sigma_compound1)
        
    else:
        
        G_compound1_cr=Compound1.kg*(sigma_compound1**2)
        B_compound1_cr=Compound1.kb*Var.M_compound1_cr*(sigma_compound1**2)
        D_compound1_cr=0
    
    #0 compound 1 crystal class size due to secondary nucleation (neglecting primary nucleation : ASSUMPTION)
    Nb_compound1_cr=B_compound1_cr*deltat
    Ng_compound1_cr=-(G_compound1_cr*Var.N_compound1_cr[0]*deltat)/deltal
    Nd_compound1_cr=-(D_compound1_cr*Var.N_compound1_cr[0]*deltat)/deltal
   
    Var.N_compound1_cr[0]=Var.N_compound1_cr[0]+Nb_compound1_cr+Ng_compound1_cr+Nd_compound1_cr-(Fout_compound1_cr*deltat*Var.N_compound1_cr[0])/Var.M_compound1_cr
    Var.V_compound1_cr_distr[0]=Var.N_compound1_cr[0]*Compound1.kv*0.25*(deltal**4)/deltal
    atot_compound1_cr=atot_compound1_cr+Var.N_compound1_cr[0]*Compound1.ka*(deltal**2)*(0**2)
    
    #1-298 crystal class size                                                                                    #The range of the loop is from 1 to classes-1=299 due to python syntax, the loop actually runs from 1 to 298
    for i in range(1,(classes-1)):
        Ng_compound1_cr=(G_compound1_cr*deltat*(Var.N_compound1_cr[i-1]-Var.N_compound1_cr[i]))/deltal
        Nd_compound1_cr=(D_compound1_cr*deltat*(Var.N_compound1_cr[i+1]-Var.N_compound1_cr[i]))/deltal

        Var.N_compound1_cr[i]=Var.N_compound1_cr[i]+Ng_compound1_cr+Nd_compound1_cr-(Var.N_compound1_cr[i]*Fout_compound1_cr*deltat)/Var.M_compound1_cr
        Var.V_compound1_cr_distr[i]=Var.N_compound1_cr[i]*Compound1.kv*0.25*((((i+1)*deltal)**4)-((i*deltal)**4))/deltal
        atot_compound1_cr=atot_compound1_cr+Var.N_compound1_cr[i]*Compound1.ka*(deltal**2)*(i**2)     
        
    #299 crystal class size
    Ng_compound1_cr=(G_compound1_cr*deltat*Var.N_compound1_cr[classes-2])/deltal
    Nd_compound1_cr=-(D_compound1_cr*deltat*Var.N_compound1_cr[classes-1])/deltal
    
    Var.N_compound1_cr[classes-1]=Var.N_compound1_cr[classes-1]+Ng_compound1_cr+Nd_compound1_cr-(Var.N_compound1_cr[classes-1]*deltat*Fout_compound1_cr)/Var.M_compound1_cr
    Var.V_compound1_cr_distr[classes-1]=Var.N_compound1_cr[classes-1]*Compound1.kv*0.25*(((classes*deltal)**4)-(((classes-1)*deltal)**4))/deltal
    atot_compound1_cr=atot_compound1_cr+Var.N_compound1_cr[classes-1]*Compound1.ka*(deltal**2)*((classes-1)**2)  
        
    #Volume difference(increase/decrease) of compound 1 crystals in this interval 
    V_compound1_cr_diff=sum(Var.V_compound1_cr_distr)-Var.V_compound1_cr_previous_loop
    
    #Mass difference(increase/decrease) of compound 1 crystals in this interval 
    M_compound1_cr_diff=V_compound1_cr_diff*Compound1.d
    
    #Mass of compound 1 crystals produced/consumed via reacting : growth and/or nucleation and/or dissolution in this time interval
    M_compound1_cr_reactions=M_compound1_cr_diff-Fin_compound1_cr*deltat+Fout_compound1_cr*deltat
    
     #%%
    
     #Function for the estimation of crystal size distribution - Part 3 :Ice crystal size distribution / Mass balance
    if DT_ice<0:                                                                                                 #If there is no ice undercooling, there is only dissolution of ice crystals
        #Ice crystals growth rate : m/s
        G_ice=0
        
        #Ice crystals nucleation rate : crystals/s 
        B_ice=0
        
        #Ice crystals dissolution rate : m/s
        D_ice=kd_ice*(-DT_ice)
        
    else:
        
        G_ice=kg_ice*DT_ice
        B_ice=kb_ice*atot_ice*(DT_ice**2)
        D_ice=0
    
    #0 ice crystal class size due to secondary nucleation (neglecting primary nucleation : ASSUMPTION)
    Nb_ice=B_ice*deltat
    Ng_ice=-(G_ice*Var.N_ice[0]*deltat)/deltal
    Nd_ice=-(D_ice*Var.N_ice[0]*deltat)/deltal

    Var.N_ice[0]=Var.N_ice[0]+Nb_ice+Ng_ice+Nd_ice-(Fout_ice*deltat*Var.N_ice[0])/Var.M_ice
    Var.V_ice_distr[0]=Var.N_ice[0]*kv_ice*0.25*(deltal**4)/deltal
    atot_ice=atot_ice+Var.N_ice[0]*ka_ice*(deltal**2)*(0**2)
     
    #1-298 crystal class size                                                                                    #The range of the loop is from 1 to classes-1=299 due to python syntax, the loop actually runs from 1 to 298
    for i in range(1,(classes-1)):
        Ng_ice=(G_ice*deltat*(Var.N_ice[i-1]-Var.N_ice[i]))/deltal
        Nd_ice=(D_ice*deltat*(Var.N_ice[i+1]-Var.N_ice[i]))/deltal
        
        Var.N_ice[i]=Var.N_ice[i]+Ng_ice+Nd_ice-(Var.N_ice[i]*Fout_ice*deltat)/Var.M_ice
        Var.V_ice_distr[i]=Var.N_ice[i]*kv_ice*0.25*((((i+1)*deltal)**4)-((i*deltal)**4))/deltal
        atot_ice=atot_ice+Var.N_ice[i]*ka_ice*(deltal**2)*(i**2)
        
    #299 crystal class size
    Ng_ice=(G_ice*deltat*Var.N_ice[classes-2])/deltal
    Nd_ice=-(D_ice*deltat*Var.N_ice[classes-1])/deltal
      
    Var.N_ice[classes-1]=Var.N_ice[classes-1]+Ng_ice+Nd_ice-(Var.N_ice[classes-1]*deltat*Fout_ice)/Var.M_ice
    Var.V_ice_distr[classes-1]=Var.N_ice[classes-1]*kv_ice*0.25*(((classes*deltal)**4)-(((classes-1)*deltal)**4))/deltal
    atot_ice=atot_ice+Var.N_ice[classes-1]*ka_ice*(deltal**2)*((classes-1)**2)
    
    #Volume difference(increase/decrease) of ice crystals in this interval
    V_ice_diff=sum(Var.V_ice_distr)-Var.V_ice_previous_loop
    
    #Mass difference(increase/decrease) of ice crystals in this interval
    M_ice_diff=V_ice_diff*d_ice
    
    #Mass of ice crystals produced/consumed via reacting : growth and/or nucleation and/or dissolution in this time interval
    M_ice_reactions=M_ice_diff-Fin_ice*deltat+Fout_ice*deltat

    return (Var)
    #%%
    
    #Function for the calculation of the mass balances for the different compounds
    
def MassBalances():    
    
    global V_ice_diff,V_compound1_cr_diff,M_ice_diff,M_compound1_cr_diff,M_ice_reactions,M_compound1_cr_reactions
    
    #Calculation of new values based on mass balances
    
    #Calculation of the new mass for every compound in the reactor
    Var.M_ice=Var.M_ice+M_ice_diff
    Var.M_compound1_cr=Var.M_compound1_cr+M_compound1_cr_diff
    #print("M_compound1_cr is "+str(Var.M_compound1_cr))
    Var.M_solids_tot=Var.M_ice+Var.M_compound1_cr
    print("Var.M_compound1 is "+str(Var.M_compound1))  
    Var.M_water=Var.M_water+Fin_water*deltat-Fout_water*deltat-M_ice_reactions                                                                
    Var.M_compound1=Var.M_compound1+Fin_compound1*deltat-Fout_compound1*deltat-M_compound1_cr_reactions   

                             
    Var.M_compound2=Var.M_compound2+Fin_compound2*deltat-Fout_compound2*deltat                                       
    Var.M_mother_liquor=Var.M_water+Var.M_compound1+Var.M_compound2             
    Var.M_liquid_tot=Var.M_mother_liquor
    
    Var.Mtot=Var.M_solids_tot+Var.M_liquid_tot
    print("Mtot in the loop is "+str(Var.Mtot))
    
    #Calculation of the new volume for every compound in the reactor
    Var.V_ice=Var.V_ice+V_ice_diff
    Var.V_compound1_cr=Var.V_compound1_cr+V_compound1_cr_diff
    Var.V_solids_tot=Var.V_ice+Var.V_compound1_cr

    Var.V_water=Var.M_water/d_water                                                                
    Var.V_compound1=Var.M_compound1/Compound1.d                                  
    Var.V_compound2=Var.M_compound2/Compound2.d                                       
    Var.V_mother_liquor=Var.V_water+Var.V_compound1+Var.V_compound2             
    Var.V_liquid_tot=Var.V_mother_liquor
    
    Var.Vtot=Var.V_solids_tot+Var.V_liquid_tot
    
    #New concentration of Na2SO4 in the bulk liquid in the reactor
    Var.C_compound1=(Var.M_compound1*1000)/(Compound1.MW*Var.M_water)                                               # Units : molals = mol compound 1/kg water
    Var.C_compound1_mliq=Var.C_compound1
    
    #Store the new volumes of ice and compound 1 crystals for the calculation of V_diff in the next time interval
    Var.V_ice_previous_loop=Var.V_ice
    Var.V_compound1_cr_previous_loop=Var.V_compound1_cr
    
    #Calculation of the new solids contents in the reactor
    Var.solids_weight_percentage=(Var.M_solids_tot/Var.Mtot)*100 
    
    return (Var)
  #%%
 
  #Function for the calculation of the heat balance 
def HeatBalances():
    global M_ice_reactions,M_compound1_cr_reactions, Qtot
    
    #Heat balance - New temperature in the reactor (!!!Assumed heat flux into the system is POSITIVE)
    Qcryst_ice=(M_ice_reactions*DHfus_ice)/deltat                                                                     #Heat flux to the system due to ice crystallization : J/s
    Qcryst_compound1_cr=(M_compound1_cr_reactions*Compound1.DHfus)/deltat                                             #Heat flux to the system due to compound 1 crystallization : J/s
    
    Qtot=Qcryst_ice+Qcryst_compound1_cr-Cooling_rate                                                                  #Total heat flux to the system : J/s
    print("Qcryst_ice is "+str(Qcryst_ice))
    print("Qcryst_compound1_cr is "+str(Qcryst_compound1_cr))
    print("Qtot is "+str(Qtot))
    Na2SO4_weight_percentage=(Var.M_compound1/Var.M_liquid_tot)*100                                                     #Na2SO4 weight percentage in the liquid in the reactor for the estimation of the heat capacity of the solution : Na2SO4 wt % = (total mass of Na2SO4 in the liquid in the reactor / total mass of liquid in the reactor)*100 %
    print("Na2SO4_weight_percentage is "+str(Na2SO4_weight_percentage))
    print("Var.M_compound1 is "+str(Var.M_compound1))
    print("Var.M_liquid_tot is "+str(Var.M_liquid_tot))

    Cp_solution=(0.9988-0.006494*Na2SO4_weight_percentage+0.00003025*(Na2SO4_weight_percentage**2)-0.0000001286*(Na2SO4_weight_percentage**3))*4.184*1000     #Heat capacity of the solution (Calculated for 25 oC, Neglecting the impurities and assuming no change of heat capacity due to temperature) (source:Heat of solution, heat capacity...) :   J/(kg*K) 
    
    Var.Treactor=(Qtot*deltat)/(Var.M_liquid_tot*Cp_solution+Var.M_compound1_cr*Compound1.Cp+Var.M_ice*Cp_ice)+Var.Treactor
    print("Treactor is "+str(Var.Treactor))
    return (Var, Qtot, Cp_solution)
#%%
        
#Function for storing several values        
def Store():
    global t_store,sigma_compound1_store,DT_ice_store,M_ice_store,M_compound1_cr_store,Treactor_store   
    t_store=np.append(t_store,Var.t)
    sigma_compound1_store=np.append(sigma_compound1_store,sigma_compound1)
    DT_ice_store=np.append(DT_ice_store,DT_ice)
    M_ice_store=np.append(M_ice_store,Var.M_ice)
    M_compound1_cr_store=np.append(M_compound1_cr_store,Var.M_compound1_cr)
    Treactor_store=np.append(Treactor_store,Var.Treactor)

 
#%%
#Function for plotting the initial values of several variables    
    
def InitialPlot():
    global graph1,graph2,graph3,graph4

    graph1=fig1.add_subplot(311)
    graph1.set_title('Number of ice crystals in each class')
    graph1.set_xlabel('Class Length (m)')
    graph1.set_ylabel('Number of ice crystals')   
    graph1.plot(class_length, Var.N_ice, 'r-',label='Initial values')
    
    graph2=fig1.add_subplot(313)
    graph2.set_title('Number of Compound 1 crystals in each class')
    graph2.set_xlabel('Class Length (m)')
    graph2.set_ylabel('Number of Comp. 1 crystals')   
    graph2.plot(class_length, Var.N_compound1_cr, 'r-',label='Initial values')
    
    graph3=fig2.add_subplot(311)
    graph3.set_title('Volume of ice crystals in each class')
    graph3.set_xlabel('Class Length (m)')
    graph3.set_ylabel('Volume of ice crystals (m^3)')   
    graph3.plot(class_length, Var.V_ice_distr, 'r-',label='Initial values')
        
    graph4=fig2.add_subplot(313)
    graph4.set_title('Volume of Compound 1 crystals in each class')
    graph4.set_xlabel('Class Length (m)')
    graph4.set_ylabel('Volume of Comp. 1 crystals (m^3)')   
    graph4.plot(class_length, Var.V_compound1_cr_distr, 'r-',label='Initial values')   
 #%%
#Function for plotting final results
    
def Plot():
    global t_store,sigma_compound1_store,DT_ice_store,M_ice_store,M_compound1_cr_store,Treactor_store,graph1,graph2,graph3,graph4 
    
    graph1.plot(class_length,Var.N_ice,'b-',label='Final values')

    graph2.plot(class_length,Var.N_compound1_cr,'b-',label='Final values')
    fig1.legend(loc='center right')
    fig1.show
    
    graph3.plot(class_length,Var.V_ice_distr,'b-',label='Final values')

    graph4.plot(class_length,Var.V_compound1_cr_distr,'b-',label='Final values')
    fig2.legend(loc='center right')
    fig2.show

    graph5=fig3.add_subplot(111)
    graph5.plot(t_store,sigma_compound1_store,'b-')
    graph5.set_title('Supersaturation of Compound 1 vs time')
    graph5.set_xlabel('Time(sec)')
    graph5.set_ylabel('Superasturation of Compound 1 (-)')
    fig3.show
    
    graph6=fig4.add_subplot(111)
    graph6.plot(t_store,DT_ice_store,'b-')
    graph6.set_title('Undercooling of ice vs time')
    graph6.set_xlabel('Time(sec)')
    graph6.set_ylabel('Undercooling of ice (K)')
    fig4.show
    
    graph7=fig5.add_subplot(311)
    graph7.set_title('Mass of ice crystals vs time')
    graph7.set_xlabel('Time(sec)')
    graph7.set_ylabel('Mass of ice crystals (kg)')
    graph7.plot(t_store,M_ice_store,'b-')
    
    graph8=fig5.add_subplot(313)
    graph8.set_title('Mass of Compound 1 crystals vs time')
    graph8.set_xlabel('Time(sec)')
    graph8.set_ylabel('Mass of Comp. 1 crystals (kg)')
    graph8.plot(t_store,M_compound1_cr_store,'b-')
    fig5.show
    
    graph9=fig6.add_subplot(111)
    graph9.set_title('Temperature in the reactor vs time')
    graph9.set_xlabel('Time(sec)')
    graph9.set_ylabel('Temperature (K)')
    graph9.plot(t_store,Treactor_store,'b-')
 #%%
# #calculate final concentration for all ions 
class conc_f:
    def __init__(self, Cna_in, Ccl_in, Ck_in, Cmg_in, Cca_in, Cso4_in, Qin, Min, Mout, M_compound1_cr2):
        self.Cna_in=Cna_in
        self.Ccl_in=Ccl_in
        self.Ck_in=Ck_in
        self.Cmg_in=Cmg_in
        self.Cca_in=Cca_in
        self.Cso4_in=Cso4_in
        self.Qin=Qin
        self.Min=Min
        self.Mout=Mout
        self.M_compound1_cr2=M_compound1_cr2
        
        
    def calc_f_c(self):
        print("self.Min is "+str(self.Min))
        print("self.Mout is "+str(self.Mout))
        print("self.M_compound1_cr2 is "+str(self.M_compound1_cr2))
        self.Ccl_out=self.Ccl_in*self.Min/self.Mout
        self.Ck_out=self.Ck_in*self.Min/self.Mout
        self.Cmg_out=self.Cmg_in*self.Min/self.Mout
        self.Cca_out=self.Cca_in*self.Min/self.Mout
        self.Mna_out=self.Cna_in*self.Min/(1000*d_in/1000)-self.M_compound1_cr2*2*constants.MW_Na/MW_Na2SO4_10H2O        
        self.Cna_out=((self.Cna_in*self.Min/(1000*d_in/1000)-self.M_compound1_cr2*2*constants.MW_Na/MW_Na2SO4_10H2O)/self.Mout)/1000#g/l
        self.Cso4_out=((self.Cso4_in*self.Min/(1000*d_in/1000)-self.M_compound1_cr2*constants.MW_so4/MW_Na2SO4_10H2O)/self.Mout)/1000#g/l
        print("Cna_out is "+str(self.Cna_out))
        print("Cso4_out is "+str(self.Cso4_out))
    
#%%
    
#Main Procedure

# Compounds()
C_Na2SO4_sol_line, tck = Compounds()
print("tck is " + str(tck))

print(C_Na2SO4_sol_line)

Var=Initialize(Var)
  
InitialPlot()

while Var.t<=55000:
        if Var.solids_weight_percentage>=40:                      #Set a maximum limit based on the solids weight percentage (wt%)
            print('Time:',Var.t,'sec,','Maximum solids weight percentage in the reactor reached : 40 wt% solids')
            break
        else:
            Var=CrystalSizeDistribution()
            Var=MassBalances()
            Var,Qtot, Cp_solution=HeatBalances()
            if ((Var.t)%100)==0:
                Store()
            else:
                pass
Plot()
print("Var.t is " +str(Var.t))
print("Var.solids_weight_percentage is " +str(Var.solids_weight_percentage))

#Calculate and print the elapsed time after the calculations
elapsed=time.time()-time0
print('Time elapsed : ',elapsed)#Calculate and print the elapsed time after the calculations
Residence_time=Var.t/3600
res=Residence_time#Vreactor/(Qf)
print(res)
print('Residence Time : ', Residence_time,'hours')
print('Volume reactor : ', Vreactor*1000,'liters')
Flow_rate=Qf/(d_in)/Residence_time
print('Continous Flow Rate : ', Flow_rate,'liters/hour')
Min=Qf#Residence_time #kg/hr 
#mass flowrate of na2so4 
M_compound1_cr=Var.M_compound1_cr*MW_Na2SO4_10H2O/MW_Na2SO4/res#Residence_time
print("mass flowrate of na2so4 "+ str(M_compound1_cr)+"kg/hr")
#mass flow rate of ice 
M_ice= Var.M_ice/res#Residence_time
print("ice mass flow rate " + str(M_ice)+ "kg/hr")
#Mass flow rate for effluent

Mout = Var.M_liquid_tot.item() / res # Residence_time
print("Mass flow rate for effluent is "+ str(Mout)+"kg/hr")


conc_f_in=conc_f(C_i_in[0],C_i_in[1], C_i_in[2], C_i_in[3], C_i_in[4], C_i_in[5],Qf/d_in, Qf, Var.M_liquid_tot.item(), M_compound1_cr)

dens=1.02
conc_f_in.calc_f_c()
Cout= Var.C_compound1_mliq #out concentration for na2so4. only na, so4 change 
Cout_so4= Cout.item()
Cout_cl=Var.C_compound2
Cout_cl= Cout_cl.item()
Cout_na= conc_f_in.Cna_out
Cout_so4= conc_f_in.Cso4_out
Cout_efc_g=[Cout_na+conc_f_in.Ccl_out/constants.MW_cl*constants.MW_Na, conc_f_in.Ccl_out, conc_f_in.Ck_out, 0, 0, Cout_so4] #g/l 
print("Cout_efc is "+str(Cout_efc_g))


M_solids_tot=Var.M_solids_tot/res#Residence_time
print("M_solids_tot is "+str(M_solids_tot))
Mtot=Var.Mtot/res#Residence_time
print("Mtot is " + str(Mtot))
M_ice=Var.M_ice/res#Residence_time
print("M_ice is "+str(M_ice))

V_compound1=Var.V_compound1/res
print("V_compound1 is "+str(V_compound1))

#operating temperature 
Treactor=Var.Treactor-273
print("operating temperature is "+str(*Treactor[0])+"oC")

#required energy 
E_el=Qtot/1000*res # KW 
Q_t=abs(Qtot/1000)
print("Q_t"+str(Q_t))
E_p=(Min*dp+Mout*dp+(M_ice+M_compound1_cr)*dp_slurry)*1e5/3600/(1000*npump)/1000 #kwh
E_fil=scaleup.scaleup(0.5, 0.3*1000, Qf)
E_t=Q_t+E_p+E_fil
print("E_t"+str(E_t))
sec_f=E_t/Qf
sec_eff=E_t/(Mout/1020)
sec_product=E_t/M_compound1_cr
w_in_cryst=M_compound1_cr*10*18.01528/MW_Na2SO4_10H2O
print(w_in_cryst)
print("M_ice*res is "+str(M_ice*res))
print("Mout*res is "+str(Mout*res))
print("M_compound1_cr*res is "+str(M_compound1_cr*res))

M_efc_out=Var.M_liquid_tot.item()
M_compound1_cr=M_compound1_cr*res
M_ice=M_ice*res





MW_Na=22.99
MW_cl=35.453
MW_so4=32.064+4*15.999
MW_K=39.102
MW_Ca=40.08
MW_Mg=24.31
MW_HCO3=1.008+12.011+3*15.999

#PRICES
#%%prices 
el_pr=0.253 #euro/kwh
s_pr=0#0.0083 #euro/kwh
co2_em_c=23 #euro/ton co2 
nacl_pr=66/1000 # euro/kg (66euro/ton)

hcl_pr=1040.0/180 #euro/l 1M solution (https://www.sigmaaldrich.com/NL/en/product/mm/100313?gclid=CjwKCAjw9pGjBhB-EiwAa5jl3EARJcEZTu2dTg76NbdyFh3fTORweHxqZex8qeIcxNPsdIFTOlF0sRoCnh4QAvD_BwE&gclsrc=aw.ds)
hcl_0_5_pr=  0#125/1000*0.5
w_pr=0.001 #euro/kg
na2so4_pr=140*0.83/1000 #euro/kg
mgoh2_pr=1000/1000 #euro/kg
caoh2_pr=125/1000#euro/kg
naoh_pr=6840/950#euro/L 1M NaOH solution  (https://www.sigmaaldrich.com/NL/en/search/1m-naoh-solution?focus=products&page=1&perpage=30&sort=relevance&term=1m%20naoh%20solution&type=product_name)
naoh_pr_s=330/1000 #euro/kg
hcl_pr_s=125/1000 #euro/kg
naoh_0_5_pr=66/1000*0.5
antisc_pr=0.002 #euro/ml
cw_pr=0.118/1000#euro/kg
m_salt_pr=5/1000#euro/kg


#%%

#Molecular weight 
MW_Na=22.99
MW_cl=35.453
MW_so4=32.064+4*15.999
MW_K=39.0983
MW_Ca=40.078
MW_Mg=24.305
MW_HCO3=1.008+12.011+3*15.999
MW_MgOH=58.3197
MW_CaOH=74.09
MW_Na2SO4=142.039458
MW_HCl=36.458
MW_NaOH=39.997
MW_NaCl=58.442729
MW_na2so410h20=322.192258

#emissions
CO2_el=0.275 # kg/kwh
CO2_st=0# kg/kwh

#input data
Csw=39.1 #g/l 

#economic
hr=24*300 #hours/year
lf= 20 #years
r=0.06 # interest rate
inf=0.02 # inflation rate 
co2_el_f=0.275 # emission factor electricity 
co2_s_f=0.633 # emission factor steam


#cost from pilot data
eq_c_nf=160000
eq_c_med=220000
eq_c_th_cr=94800
eq_c_mfpfr=80000
eq_c_efc=72579.59
eq_c_edbm= 90000   
eq_c_ev_pd=10000
eq_c_nf_conc=7000
eq_c_ed=123437.5

#capacity of basic scenario
#Mf_basic_sc=nf, med, thermal cryst, mfpfr, nf conc step, efc, edbm, ev_ponds ]
Mf_basic_sc=[2290.858979,1695.830673,199.5094909, 595.0283063,300,100, 229.0522085, 253, 55130.92]


import math
from desalsim.density_calc import density_calc 
from desalsim import constants 
from desalsim import scaleup

#%%
    #Molecular weight 
MW_Na=constants.MW_Na
MW_cl=constants.MW_cl
MW_so4=constants.MW_so4
MW_K=constants.MW_K
MW_Ca=constants.MW_Ca
MW_Mg=constants.MW_Mg
MW_HCO3=constants.MW_HCO3
MW_NaCl=constants.MW_NaCl
MW_KCl=MW_K+MW_cl
MW_Mgso4=MW_Mg+MW_so4
MW_Caso4=MW_Ca+MW_so4
MW_Na2so4=2*MW_Na+MW_so4
MW_MgOH=constants.MW_MgOH

#%%

class thermal_calc:
    """
    This class calculates the mass and energy balance for thermal crystallization.

    Attributes
    ----------
        T_op : float 
            Operating temperature (oC)
        Qf : float
            Inlet flow rate (kg/hr)
        Cf_s : float
        Total ion concentration of solution (g/L)
        Cf_caso4 : float
            Concentration of CaSO4 in the solution (g/L)
        T_in : float
            Inlet feed temperature (oC)
        d_sol : float
            Inlet feed density (kg/L) 
        Cf_in : float
            List of concentration of ions in the solution (g/L)
        nacl_sat_wt : float
            Weight percent of NaCl at saturation
        nacl_sol : float
            NaCl in the solution (kg/h)
        salt_solids : float
            Salt solids (kg/h)
        caso4_sol : float
            CaSO4 in the solution (kg/h)
        solid_mass : float
            Solid mass (kg/h)
        ev_mass : float
            Evaporated mass (kg/h)
        BPE : float
            Boiling point elevation (oC)
        heat_req : float
            Heat required (kJ/h)
        T_s : float
            Steam temperature (oC)
        steam_mass : float
            Steam mass (kg/h)
        cw_mass : float
            Cooling water mass (kg/h)
    """
    def __init__(self, T_op, Qf, Cf_s, Cf_caso4, T_in, Cf_in, salt_mois, LHV_v, LHV_s, T_cw_o, T_cw_f ):
        self.T_op = T_op 
        self.Qf = Qf 
        self.Cf_s = Cf_s 
        self.Cf_caso4 = Cf_caso4
        self.T_in = T_in
        self.Cf_in = Cf_in
        self.nacl_sat_wt = salt_mois
        self.d_sol=density_calc(T_in, Cf_s)/1000
        self.LHV_v=LHV_v
        self.LHV_s =LHV_s 
        self.T_cw_o=T_cw_o
        self.T_cw_f=T_cw_f
        
    def mass_bal_cryst(self):
        """
        This method calculates the mass balance for thermal crystallization.
        """
        self.nacl_sol = self.Qf*self.Cf_in[0]/self.d_sol/1000*MW_NaCl/MW_Na
        self.salt_solids = self.Qf*self.Cf_s/self.d_sol/1000 #kg/h
        self.caso4_sol = self.Qf*self.Cf_caso4/100 #kg/h
        self.solid_mass = (self.salt_solids)/(1-self.nacl_sat_wt/100) #kg/h
        self.ev_mass = self.Qf-self.solid_mass #kg/h

    def heat_bal_cryst(self):
        """
        This method calculates the heat balance for thermal crystallization.
        """
        Cp_f=3.14 # Feed specific heat capacity (units: KJ* Kg*oC)
        CP_cw=4.187 # Water specific heat capacity (units: KJ* Kg*oC)
        UA=45990
        self.BPE = 0.0127*(self.nacl_sat_wt/10)**2+0.1743*(self.nacl_sat_wt/10+0.3443)*5/9 #oC
        self.heat_req = self.LHV_v*self.ev_mass+self.Qf*Cp_f*(self.T_op-self.T_in) #kJ/h
        self.T_s = self.heat_req/UA+self.T_op #steam temperature oC
        self.steam_mass = self.heat_req/self.LHV_s 
        self.cw_mass = (self.ev_mass*self.LHV_v)/(CP_cw*(self.T_cw_o-self.T_cw_f)) 


# This class calculates the concentration of different ions in the solution
class conc_cal:
    """
    A class used to calculate the concentration of different ions in a solution.

    Attributes
    ----------
    Qf : float
        Flow rate (units: m^3/h)
    solid_mass : float
        Mass of solid (units: kg)
    Cc1, Cc2, Cc3, Cc4, Cc5, Cc6 : float
        Concentrations of six different ions (units: ppm)
    d_sol : float
        Density of the solution at 40°C (units: kg/m^3)

    Methods
    -------
    molarity():
        Calculates the molarity of different ions in the solution (units: mol/L)
    conc_salt_comp():
        Calculates the concentration of different salts in the solution (units: ppm)
    salt_conc():
        Calculates the concentration of different salts in the outlet stream (units: ppm)
    """
    def __init__(self, Qf, solid_mass, C1, Cc1, C2, Cc2, C3, Cc3, C4, Cc4, C5, Cc5, C6, Cc6,T_in):
        self.Qf=Qf
        self.solid_mass=solid_mass
        self.C1=C1
        self.C2=C2
        self.C3=C3
        self.C4=C4
        self.C5=C5
        self.C6=C6
        self.cf_in=Cc1+Cc2+ Cc3+ Cc4+ Cc5+ Cc6
        self.Cc1=Cc1
        self.Cc2=Cc2
        self.Cc3=Cc3
        self.Cc4=Cc4
        self.Cc5=Cc5
        self.Cc6=Cc6
        # Calculate the density of the feed solution at T_in 
        self.d_sol=density_calc(T_in, self.cf_in)/1000
    
    
    def molarity(self):        
        """This method calculates the molarity of different ions in the solution"""
        self.Cc1_mol=self.Cc1/MW_Na
        self.Cc2_mol=self.Cc2/MW_cl
        self.Cc3_mol=self.Cc3/MW_K
        self.Cc4_mol=self.Cc4/MW_Mg
        self.Cc5_mol=self.Cc5/MW_Ca
        self.Cc6_mol=self.Cc6/MW_so4
        self.CNacl_mol=self.Cc2_mol-self.Cc3_mol
        self.CKCl_mol=self.Cc3_mol
        self.CMgoh_mol=self.Cc4_mol
        self.CCaso4_mol=self.Cc6_mol
        self.CCaoh_mol=self.Cc6_mol-self.Cc5_mol
    
    def conc_salt_comp(self):
        """This method calculates the concentration of different salts in the solution"""
        self.CNacl=self.CNacl_mol*MW_NaCl
        self.CKCl=self.CKCl_mol*MW_KCl
        self.CMgoh=self.CMgoh_mol*MW_MgOH
        self.CCaso4=self.CCaso4_mol*MW_Caso4
        
    def salt_conc(self):
        """This method calculates the concentration of different salts in the outlet stream"""
        self.CNacl_out=self.CNacl/(self.solid_mass) 
        self.CKCl_out=self.CKCl/(self.solid_mass)
        self.CMgoh_out=self.CMgoh/(self.solid_mass)
        self.CCaso4_out=self.CCaso4/(self.solid_mass)
        self.CNa=self.Cc1*self.Qf/(self.solid_mass) #g/kg
        self.CCl=self.Cc2*self.Qf/(self.solid_mass) #g/kg
        self.CK=self.Cc3*self.Qf/(self.solid_mass) #g/kg
        self.CCa=self.Cc5*self.Qf/(self.solid_mass) #g/kg
        self.CSO4=self.Cc6*self.Qf/(self.solid_mass) #g/kg
        self.CMg=self.Cc4*self.Qf/(self.solid_mass) #g/kg


def calculate_energy(Qf, Q_evap_mass, Qcw, M_Nacl, heat_req, d_sol, dp_f, dp_w, dp_slurry, dp_cw, npump):
    """
    Calculate the energy required for thermal crystallization.

    Arguments 
    ----------
        Qf : float
            Flow rate of the feed solution in m3/h.
        Q_evap_mass : float
            Distillate flow rate m3/h.
        Qcw : float
            Flow rate of cooling water in m3/h.
        d_sol : float
            Density of the feed solution in kg/m3.
        npump : float
            Pump efficiency.

    Returns:
        tuple: A tuple containing the following:
            SEC_el : float
                Specific energy consumption for electrical energy in kWh/m3.
            SEC_el_prod : float
                Specific energy consumption for electrical energy per unit mass of evaporated water in kWh/kg.
            SEC_el_prod2 : float
                Specific energy consumption for electrical energy per unit mass of NaCl produced in kWh/kg.
    """
    
    # Calculate electrical energy consumption
    E_el_th_Cr = ((Qf*dp_f + Q_evap_mass*dp_w + M_Nacl*dp_slurry + 2*Qcw*dp_cw)*1e5/3600/(1000*npump))/1000 #kwh
    
    # Calculate thermal energy consumption
    E_th_th_Cr = heat_req/3600 #kwh
    
    # Calculate energy consumption for filtration
    E_fil = 0
    
    # Add filtration energy consumption to electrical energy consumption
    E_el_th_Cr = E_el_th_Cr + E_fil
    
    # Calculate specific electrical energy consumption per unit mass of feed 
    SEC_el_f = (E_fil + E_el_th_Cr)/(Qf*d_sol*1000)
    
    # Calculate specific electrical energy consumption per unit mass of evaporated water
    SEC_el_w = (E_fil + E_el_th_Cr)/(Q_evap_mass/1000)
    
    # Calculate specific electrical energy consumption per unit mass of NaCl produced
    SEC_el_NaCl = (E_fil + E_el_th_Cr)/(M_Nacl)
    
    return round(E_el_th_Cr,2), round(E_th_th_Cr,2), round(SEC_el_f,2), round(SEC_el_NaCl,2), round(SEC_el_w,2)

def mass_balance(Qf, Cf_in, M_Nacl, Csalt_out, d_sol):
    """Calculate the mass balance."""
    for i in range(len(Cf_in)):
        bal_i=Cf_in[i]*(Qf)/1000 - (M_Nacl*Csalt_out[i]/1000)
        print(f"For {i}, mass balance equals: {bal_i}")

 


from desalsim.density_calc import density_calc 

#%%calculations 
class MEDCalculator:
    """
    A class used to represent Mass Balance for Multi-Effect Distillation 

    ...

    Attributes
    ----------
    Qf : float
        Flow rate (m^3/s).
    CNa_in, CCl_in, CK_in, CMg_in, CCa_in, CSO4_in : float
        Initial concentrations of various ions (g/l).

    Methods
    -------
    salinity_calc():
        Calculates the inflow salinity.
        
    mass_balance_med():
        Performs mass balance calculations.
        
    temperature_calc(T_s, T_N, T_cw_in, DT_loss):
        Calculates temperature-related parameters.
        
    performance_parameters():
        Calculates performance parameters.
        
    output_concentration():
        Calculates the concentrations in the output.
 
    """
    
    def __init__(self,Qf, Mf_med, CNa_in, CCl_in, CK_in, CMg_in, CCa_in, CSO4_in, T_in):
        #Initialize class attributes
        self.Qf = Qf
        self.Mf_med=Mf_med
        self.CNa_in = CNa_in
        self.CCl_in = CCl_in
        self.CK_in = CK_in
        self.CMg_in = CMg_in
        self.CCa_in = CCa_in
        self.CSO4_in = CSO4_in
        self.cons=[self.CNa_in,self.CCl_in,self.CK_in,self.CMg_in,self.CCa_in,self.CSO4_in]
        self.T_in=T_in
        self.d=density_calc(self.T_in, sum(self.cons))


    def salinity_calc(self):
        # Calculate inflow salinity
        self.salinity_in=sum(self.cons)
        self.xf=self.salinity_in/self.d*1000
        self.Mf=self.Mf_med/3600 #kg/s
        
        
    def mass_balance_med(self, Cb_out):
        # Perform mass balance calculations
        #Brine Concentration leaving effect n 
        self.xn=Cb_out
        
        #Calculate concentration factor 
        self.conc_f=self.xn/self.xf
        
        #Calculate brine flow rate of leaving effect n
        self.Bn=self.xf*self.Mf/self.xn #kg/s
        self.Qb=self.Bn*3600 #kg/hr
        
        #Calculate total distillate flow rate
        self.Mdist=self.Mf-self.Bn #kg/s
        self.Qdist=self.Mdist*3600 #kg/hr
        
        #Calculate distillate flow produces for each effect
        self.D1=self.Mdist/(1+self.lhv1/self.lhv2) #kg/s
        self.D2=self.D1*self.lhv1/self.lhv2 #kg/s

        
        
    def temperature_calc(self, DT_loss, T_N, T_s):
        # Calculate temperature-related parameters
        T3=69
        #Transfer coefficients of the effects,
        self.U1= 1.9695+0.012057*T3-0.000085989*T3**2+0.00000025651*T3**3
        self.U2=0.95*self.U1
        
        #Temperature drop in each effect 
        self.DT_t=T_s-T_N #oC
        self.DT1=self.DT_t/(self.U1*(1/self.U1+1/self.U2)) #oC
        self.DT2=self.DT1*self.U1/self.U2 #oC
        
        #Temperature profile of each effect 
        self.T1=T_s-self.DT1 #oC
        self.T2=self.T1-self.DT2 #oC
        
        #Latent heat of first effect 
        self.lhv1=2499.5698-0.9156*(self.T1-DT_loss)-0.048343*(self.T1-DT_loss)**2
        
        #Latent heat of second effect
        self.lhv2=2626.1
    
    def performance_parameters(self, lh_s,  T_cw_in):
        # Calculate performance parameters
        #Steam mass flow rate
        self.Ms=self.D1*self.lhv1/lh_s #kg/s
        
        #Gain Output Ratio
        self.GOR=self.Mdist/self.Ms
        
        #Condenser thermal load (Qc) kj/s ot kw
        self.Qc=self.D2*self.lhv2 
        
        #thermal load in the first effect (Q1) kw
        self.Q1= self.Ms*lh_s
        
        #Cooling water flow rate kg/s
        Cp_w=4182 # specific heat capacity of water (j/kgC)
        self.QCW= self.D2*self.lhv2*1000/(Cp_w*(self.T_in-T_cw_in))-self.Mf
        
        #Performance ratio (PR) 
        self.PR= self.Mdist/self.Mf
        
        #Condenser thermal load
        cp_sol=4184 # specific heat capacity of solution (j/kgC)
        self.Qsen=self.Mf*cp_sol*(self.T_in-25)
        
        #Total thermal load
        self.Q_Tot=self.Q1+self.Qsen/1000

    def output_concentration(self):
        # Calculate concentrations in the output, g/l
        self.CNa_out=self.CNa_in*self.Mf/self.Bn 
        self.CCl_out=self.CCl_in*self.Mf/self.Bn 
        self.CK_out=self.CK_in*self.Mf/self.Bn 
        self.CMg_out=self.CMg_in*self.Mf/self.Bn 
        self.CCa_out=self.CCa_in*self.Mf/self.Bn  
        self.CSO4_out=self.CSO4_in*self.Mf/self.Bn  
